from abc import ABC, abstractmethod
import logging
import os
import warnings

import six
import numpy as np
import pyarrow as pa

import vaex
import vaex.utils
import vaex.cache
from .array_types import supported_array_types, supported_arrow_array_types, string_types, is_string_type


if vaex.utils.has_c_extension:
    import vaex.strings

logger = logging.getLogger("vaex.column")


register = vaex.encoding.make_class_registery('column')

class Column(ABC):
    _cached_fingerprint = None

    def tolist(self):
        return self.to_numpy().tolist()

    def fingerprint(self):
        '''id that uniquely identifies a column cross runtime'''
        if self._cached_fingerprint is None:
            self._cached_fingerprint = self._fingerprint()
        return self._cached_fingerprint

    @abstractmethod
    def _fingerprint(self):
        raise NotImplementedError

    def to_arrow(self, type=None):
        return pa.array(self, type=type)

supported_column_types = (Column, ) + supported_array_types

def is_column_like(col):
    return isinstance(col, supported_column_types)


class ColumnVirtualRange(Column):
    def __init__(self, start, stop, step=1, dtype=None):
        self.start = start
        self.stop = stop
        self.step = step
        self.dtype = np.dtype(dtype)
        self.shape = (self.__len__(),)

    def __len__(self):
        return (self.stop - self.start) // self.step

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.start, self.stop, self.step, self.dtype, self.shape)
        return f'column-vrange-{fp}'

    def __getitem__(self,  slice):
        start, stop, step = slice.start, slice.stop, slice.step
        if start is None:
            start = self.start
        if stop is None:
            stop = self.stop
        if step is None:
            step = self.step
        return np.arange(self.start + start, self.start + stop, step, dtype=self.dtype)

    def trim(self, i1, i2):
        return ColumnVirtualRange(self.start + i1 * self.step, self.start + i2 * self.step, self.step, self.dtype)


class ColumnVirtualConstant(Column):
    def __init__(self, value, length, dtype=None, chunk_size=1024):
        self.value = value
        self.length = length
        self.dtype = self._get_dtype(dtype)
        self.chunk_size = chunk_size

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        template = pa.array([self.value] * self.chunk_size, type=self.dtype.arrow)
        n_chunks = (self.length + self.chunk_size - 1) // self.chunk_size
        if item.step not in [None, 1]:
            raise ValueError('step can only be 1')
        start, stop, step = item.start, item.stop, item.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        item = slice(start, stop, step)
        ar = pa.chunked_array([template] * n_chunks)[item]
        return ar

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.value, self.length, self.dtype, self.chunk_size)
        return f'column-vconstant-{fp}'

    def _get_dtype(self, dtype):
        if dtype is None:
            return vaex.datatype.DataType(pa.array([self.value]).type)
        else:
            return vaex.datatype.DataType(dtype)

    def trim(self, i1, i2):
        length = i2 - i1
        return ColumnVirtualConstant(value=self.value, length=length)


class ColumnMaskedNumpy(Column):
    def __init__(self, data, mask):
        self.data = data
        self.mask = mask
        self.dtype = data.dtype
        assert len(data) == len(mask)

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.data[:])
        hash2 = vaex.dataset.hash_array_data(self.indimaskces[:])
        fp = vaex.cache.fingerprint(hash1, hash2)
        return f'column-masked-numpy-{fp}'

    def __len__(self):
        return len(self.data)

    def __getitem__(self,  slice):
        data = self.data[slice]
        mask = self.mask[slice]
        return np.ma.array(data, mask=mask, shrink=False)

    def trim(self, i1, i2):
        return ColumnMaskedNumpy(self.data.trim(i1, i2), self.mask.trim(i1, i2))


class ColumnSparse(Column):
    def __init__(self, matrix, column_index):
        self.matrix = matrix
        self.column_index = column_index
        self.shape = self.matrix.shape[:1]
        self.dtype = self.matrix.dtype

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.matrix, self.column_index)
        return f'column-sparse-{fp}'

    def __len__(self):
        return self.shape[0]

    def trim(self, i1, i2):
        return ColumnSparse(self.matrix[i1:i2], column_index=self.column_index)

    def __getitem__(self, slice):
        # not sure if this is the fastest
        return self.matrix[slice, self.column_index].A[:,0]


class ColumnNumpyLike(Column):
    """Wraps a numpy like object (like hdf5 dataset) into a column vaex understands"""
    def __init__(self, ar):
        self.ar = ar  # this should behave like a numpy array
        self.dtype = ar.dtype

    def __len__(self):
        return len(self.ar)

    def trim(self, i1, i2):
        return type(self)(self.ar[i1:i2])

    def __getitem__(self, slice):
        return self.ar[slice]

    def __setitem__(self, slice, value):
        self.ar[slice] = value

    def _fingerprint(self):
        ar = self[:]
        hash = vaex.dataset.hash_array(ar)
        fp = vaex.cache.fingerprint(hash)
        return f'column-numpy-like-{fp}'


@register
class ColumnArrowLazyCast(Column):
    snake_name = "lazy_cast"
    """Wraps an array like object and cast it lazily"""
    def __init__(self, ar, type):
        self.ar = ar  # this should behave like a numpy array
        self.type = vaex.dtype(type)

    def encode(self, encoding):
        return {
            'array': encoding.encode('array', self.ar),
            'dtype': encoding.encode('dtype', self.type),
        }

    @classmethod
    def decode(cls, encoding, spec):
        return cls(encoding.decode('array', spec['array']),
                   encoding.decode('dtype', spec['dtype']))

    def __arrow_array__(self, type=None):
        return pa.array(self.ar)

    def _fingerprint(self):
        hash = vaex.dataset.hash_array_data(self.ar)
        fp = vaex.cache.fingerprint(hash, self.type)
        return f'column-arrow-lazy-cast-{fp}'

    @property
    def nbytes(self):
        return self.ar.nbytes

    @property
    def dtype(self):
        return self.type

    def __len__(self):
        return len(self.ar)

    def trim(self, i1, i2):
        return type(self)(self.ar[i1:i2], self.type)

    def __getitem__(self, slice):
        if self.ar.dtype == object and vaex.array_types.is_string_type(self.type):
            # this seem to be the only way to convert mixed str and nan to include nulls
            return pa.Array.from_pandas(self.ar[slice], type=self.type.arrow)
        return pa.array(self.ar[slice], type=self.type.arrow)


@register
class ColumnArrowDictionaryEncoded(Column):
    snake_name = "dictionary_encoded"
    """Wraps an array like object and cast it lazily to arrow dict encoded"""
    def __init__(self, indices, dictionary):
        self.indices = indices  # should be column like
        self.dictionary = vaex.array_types.to_arrow(dictionary)  # lets do this once for performance

    def encode(self, encoding):
        return {
            'indices': encoding.encode('array', self.indices),
            'dictionary': encoding.encode('array', self.dictionary),
        }

    @classmethod
    def decode(cls, encoding, spec):
        return cls(encoding.decode('array', spec['indices']),
                   encoding.decode('array', spec['dictionary']))

    def __arrow_array__(self, type=None):
        return self[:]

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.indices)
        hash2 = vaex.dataset.hash_array_data(self.dictionary)
        fp = vaex.cache.fingerprint(hash1, hash2)
        return f'column-arrow-{self.snake_name}-{fp}'

    @property
    def nbytes(self):
        # consistent with arrow
        if int(pa.__version__.split(".")[0]) >= 7:
            return self.indices.nbytes + self.dictionary.nbytes
        else:
            return self.indices.nbytes

    @property
    def dtype(self):
        return self[0:0].type

    def __len__(self):
        return len(self.indices)

    def trim(self, i1, i2):
        return type(self)(self.indices[i1:i2], self.dictionary)

    def __getitem__(self, slice):
        return pa.DictionaryArray.from_arrays(self.indices[slice], self.dictionary, safe=False)


class ColumnIndexed(Column):
    def __init__(self, column, indices, masked=False):
        self.column = column
        self.indices = indices
        self.dtype = vaex.array_types.data_type(column)
        self.shape = (len(indices),)
        self.masked = masked
        # this check is too expensive
        # max_index = self.indices.max()
        # if not np.ma.is_masked(max_index):
        #     assert max_index < self.df._length_original

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.column[:])
        hash2 = vaex.dataset.hash_array_data(self.indices[:])
        fp = vaex.cache.fingerprint(hash1, hash2, self.masked)
        return f'column-indexed-{fp}'

    @staticmethod
    def index(column, indices, direct_indices_map=None, masked=False):
        """Creates a new column indexed by indices which avoids nested indices

        :param df: Dataframe where column comes from
        :param column: Column object or numpy array
        :param name: name of column
        :param indices: ndarray with indices which rows to take
        :param direct_indices_map: cache of the nested indices (pass a dict for better performance), defaults to None
        :return: A column object which avoids two levels of indirection
        :rtype: ColumnIndexed
        """
        direct_indices_map = direct_indices_map if direct_indices_map is not None else {}
        if isinstance(column, ColumnIndexed):
            if id(column.indices) not in direct_indices_map:
                direct_indices = column.indices[indices]
                if masked:
                    # restored sentinel mask values
                    direct_indices[indices == -1] = -1
                direct_indices_map[id(column.indices)] = direct_indices
            else:
                direct_indices = direct_indices_map[id(column.indices)]
            return ColumnIndexed(column.column, direct_indices, masked=masked or column.masked)
        else:
            return ColumnIndexed(column, indices, masked=masked)

    def __len__(self):
        return len(self.indices)

    def trim(self, i1, i2):
        return ColumnIndexed(self.column, self.indices[i1:i2], masked=self.masked)

    def __arrow_array__(self, type=None):
        # TODO: without a copy we get a buserror
        # TODO2: weird, this path is not triggered anymore
        # values = self[:]
        # if hasattr(values, "to_arrow"):
        #     return values.to_arrow()
        # else:
        return pa.array(self)

    def to_numpy(self):
        return np.array(self[:])

    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        indices = self.indices[start:stop]
        ar_unfiltered = self.column
        if self.masked:
            mask = indices == -1
        if isinstance(ar_unfiltered, Column):
            # TODO: this is a workaround, since we do not require yet
            # that Column classes know how to deal with indices, we get
            # the minimal slice, and get those (not the most efficient)
            if self.masked:
                unmasked_indices = indices[~mask]
                if len(unmasked_indices) > 0:
                    i1, i2 = np.min(unmasked_indices), np.max(unmasked_indices)
                else:
                    i1, i2 = 0, 0
            else:
                i1, i2 = np.min(indices), np.max(indices)
            ar_unfiltered = ar_unfiltered[i1:i2+1]
            if self.masked:
                indices = indices - i1
            else:
                indices = indices - i1
        take_indices = indices
        if self.masked:
            # arrow and numpy do not like the negative indices, so we set them to 0
            take_indices = indices.copy()
            take_indices[mask] = 0
        if isinstance(ar_unfiltered, supported_arrow_array_types):
            ar = ar_unfiltered.take(vaex.array_types.to_arrow(take_indices))
        else:
            ar = ar_unfiltered[take_indices]
        assert not np.ma.isMaskedArray(indices)
        if self.masked:
            # TODO: we probably want to keep this as arrow array if it originally was
            return np.ma.array(ar, mask=mask)
        else:
            return ar


class ColumnConcatenatedLazy(Column):
    def __init__(self, expressions, dtype=None, shape=None, fill_value=None):
        self.is_masked = any([e.is_masked for e in expressions])
        self.fill_value = fill_value
        if self.is_masked and fill_value is None:
            for expression in expressions:
                if expression.is_masked:
                    try:
                        # fast path
                        self.fill_value = expression[0:1].fill_value
                        break
                    except:  # noqa
                        # slower path (we have to evaluate everything)
                        self.fill_value = expression.values.fill_value
                        break
            else:
                raise ValueError('Concatenating expressions with masked values, but no fill value is found')
        if dtype is None:
            dtypes = [e.dtype for e in expressions]

            any_strings = any([is_string_type(dtype) for dtype in dtypes])
            if any_strings:
                self.dtype = pa.string()  # TODO: how do we know it should not be large_string?
            else:
                # np.datetime64/timedelta64 and find_common_type don't mix very well
                if all([dtype == 'datetime64' for dtype in dtypes]):
                    self.dtype = dtypes[0]
                elif all([dtype == 'timedelta64' for dtype in dtypes]):
                    self.dtype = dtypes[0]
                else:
                    if all([dtype == dtypes[0] for dtype in dtypes]):  # find common types doesn't always behave well
                        self.dtype = dtypes[0]
                    if  any([dtype.kind in 'SU' for dtype in dtypes]):  # strings are also done manually
                        if all([dtype.kind in 'SU' for dtype in dtypes]):
                            index = np.argmax([dtype.itemsize for dtype in dtypes])
                            self.dtype = dtypes[index]
                        else:
                            index = np.argmax([df.columns[self.column_name].astype('O').astype('U').dtype.itemsize for df in dfs])
                            self.dtype = dfs[index].columns[self.column_name].astype('O').astype('U').dtype
                    else:
                        self.dtype = np.find_common_type([k.numpy for k in dtypes], [])
                    logger.debug("common type for %r is %r", dtypes, self.dtype)
            # make sure all expression are the same type
            self.expressions = [e if vaex.array_types.same_type(e.dtype, self.dtype) else e.astype(self.dtype) for e in expressions]
        else:
            # if dtype is given, we assume every expression/column is the same dtype
            self.dtype = dtype
            self.expressions = expressions[:]
        if shape is not None:
            self.shape = (len(self),) + shape
        else:
            self.shape = (len(self), ) + self.expressions[0].evaluate(0, 1, array_type='numpy', parallel=False).shape[1:]
            for i in range(1, len(self.expressions)):
                expression = self.expressions[i]
                shape_i = (len(self), ) + expressions[i].evaluate(0, 1, array_type='numpy', parallel=False).shape[1:]
                if self.shape != shape_i:
                    raise ValueError("shape of of expression %s, array index 0, is %r and is incompatible with the shape of the same column of array index %d, %r" % (self.expressions[0], self.shape, i, shape_i))

    def _fingerprint(self):
        fp = vaex.cache.fingerprint([k.fingerprint() for k in self.expressions])
        return f'column-concat-{fp}'

    def to_arrow(self, type=None):
        values = [e.values for e in self.expressions]
        chunks = [value if isinstance(value, pa.Array) else pa.array(value, type=type) for value in values]
        types = [chunk.type for chunk in chunks]

        # upcast if mixed types
        if pa.string() in types and pa.large_string() in types:
            def _arrow_string_upcast(array):
                if array.type == pa.large_string():
                    return array
                if array.type == pa.string():
                    import vaex.arrow.convert
                    column = vaex.arrow.convert.column_from_arrow_array(array)
                    column.indices = column.indices.astype(np.int64)
                    return pa.array(column)
                else:
                    raise ValueError('Not a string type: %r' % array)
            chunks = [_arrow_string_upcast(chunk) for chunk in chunks]

        return pa.chunked_array(chunks)

    def __len__(self):
        return sum(len(e.df) for e in self.expressions)

    def trim(self, i1, i2):
        start, stop = i1, i2
        i = 0  # the current expression
        offset = 0  # and the offset the current expression has wrt the dataframe
        # find the first expression we overlap with
        while start >= offset + len(self.expressions[i].df):
            offset += len(self.expressions[i].df)
            i += 1
        # do need more than 1?
        if stop > offset + len(self.expressions[i].df):
            # then add the first one
            expressions = [self.expressions[i][start-offset:]]
            offset += len(self.expressions[i].df)
            # we need more expression
            i += 1
            # keep adding complete expression till we find the it to be the last
            while stop > offset + len(self.expressions[i].df):
                offset += len(self.expressions[i].df)
                expressions.append(self.expressions[i])
                i += 1
            if stop > offset:  # add the tail part
                expressions.append(self.expressions[i][:stop-offset])
        else:
            # otherwise we only need a slice of the first
            expressions = [self.expressions[i][start-offset:stop-offset]]

        return ColumnConcatenatedLazy(expressions, self.dtype, self.shape[1:], fill_value=self.fill_value)

    def __arrow_array__(self, type=None):
        return pa.array(self[:], type=type)

    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        dtype = self.dtype
        if is_string_type(dtype):
            dtype = 'O'  # we store the strings in a dtype=object array
        expressions = iter(self.expressions)
        current_expression = next(expressions)
        offset = 0
        while start >= offset + len(current_expression.df):
            offset += len(current_expression.df)
            current_expression = next(expressions)
        # this is the fast path, no copy needed
        if stop <= offset + len(current_expression.df):
            if current_expression.df.filtered:  # TODO this may get slow! we're evaluating everything
                warnings.warn("might be slow, you have concatenated expressions with a filter set")
            values = current_expression.evaluate(i1=start - offset, i2=stop - offset, parallel=False)
        else:
            if self.is_masked:
                copy = np.ma.empty(stop - start, dtype=dtype)
                copy.fill_value = self.fill_value
            else:
                copy = np.zeros(stop - start, dtype=dtype)
            copy_offset = 0
            while offset < stop:  # > offset + len(current_expression):
                if current_expression.df.filtered:  # TODO this may get slow! we're evaluating everything
                    warnings.warn("might be slow, you have concatenated DataFrames with a filter set")
                part = current_expression.evaluate(i1=start-offset, i2=min(len(current_expression.df), stop - offset), parallel=False)
                copy[copy_offset:copy_offset + len(part)] = part
                offset += len(current_expression.df)
                copy_offset += len(part)
                start = offset
                if offset < stop:
                    current_expression = next(expressions)
            values = copy
        if is_string_type(dtype):
            return _to_string_column(values)
        else:
            return values


use_c_api = True


class ColumnString(Column):
    pass


class ColumnStringArray(Column):
    """Wraps a numpy array with dtype=object, containing all strings"""
    def __init__(self, array):
        # assert vaex.strings.all_strings(array, allow)
        self.array = array
        self.string_sequence = vaex.strings.StringArray(self.array)

    def __getitem__(self, slice):
        return ColumnString(self.array[slice])

    def to_numpy(self):
        return self.array

def _to_string(string_bytes):
    if six.PY2:
        return unicode(buffer(string_bytes))
    else:
        return bytes(memoryview(string_bytes)).decode('utf8')


def _is_stringy(x):
    if vaex.array_types.is_string(x):
        return True
    elif isinstance(x, ColumnString):
        return True
    elif isinstance(x, np.ndarray):
        if x.dtype.kind in 'US':
            return True
    return False


def _to_string_sequence(x, force=True):
    if isinstance(x, pa.DictionaryArray):
        x = x.dictionary.take(x.indices)  # equivalent to PyArrow 5.0.0's dictionary_decode() but backwards compatible
    if isinstance(x, pa.ChunkedArray):
        # turn into pa.Array, TODO: do we want this, this may result in a big mem copy
        table = pa.Table.from_arrays([x], ["single"])
        table_concat = table.combine_chunks()
        column = table_concat.columns[0]
        if column.num_chunks == 0:
            x = pa.array([], type=column.type)
        else:
            assert column.num_chunks == 1
            x = column.chunk(0)

    if isinstance(x, ColumnString):
        return x.string_sequence
    elif isinstance(x, pa.Array):
        from vaex.arrow import convert
        return convert.column_from_arrow_array(x).string_sequence
    elif isinstance(x, np.ndarray):
        mask = None
        if np.ma.isMaskedArray(x):
            mask = np.ma.getmaskarray(x)
            x = x.data
        if x.dtype == 'O':
            return vaex.strings.StringArray(x) if mask is None else vaex.strings.StringArray(x, mask)
        elif x.dtype.kind in 'US':
            x = x.astype('O')
            return vaex.strings.StringArray(x) if mask is None else vaex.strings.StringArray(x, mask)
        else:
            # This path is only required because the str_pandas wrapper uses NaN for missing values
            # see pandas_wrapper in functions.py
            if force:
                length = len(x)
                bytes = np.zeros((0,), dtype=np.uint8)
                indices = np.zeros((length+1,), dtype=np.int64)
                null_bitmap = np.ones(((length + 7) // 8,), dtype=np.uint8)
                return vaex.strings.StringList64(bytes, indices, length, 0, null_bitmap, 0)
            else:
                ValueError('unsupported dtype ' +str(x.dtype))
    elif isinstance(x, (vaex.superstrings.StringList32, vaex.superstrings.StringList64)):
        return x
    elif isinstance(x, (list, type)):
        return _to_string_sequence(np.array(x))
    else:
        raise ValueError('not a ColumnString or ndarray: ' + str(x))

def _to_string_column(x, force=True):
    ss = _to_string_sequence(x, force=force)
    if not isinstance(ss, (vaex.strings.StringList64, vaex.strings.StringList32)):
        ss = ss.to_arrow()
    return ColumnStringArrow.from_string_sequence(ss)

def _to_string_list_sequence(x):
    if isinstance(x, vaex.strings.StringListList):
        return x
    else:
        raise ValueError('not a StringListList')

def _asnumpy(ar):
    if isinstance(ar, np.ndarray):
        return ar
    else:
        return ar.to_numpy()


def _trim(column, i1, i2):
    if isinstance(column, np.ndarray):
        return column[i1:i2]
    else:
        return column.trim(i1, i2)

def _trim_bits(column, i1, i2):
    i1_bytes = i1 // 8
    offset = i1 - i1_bytes * 8
    i2_bytes = (i2 + 7) // 8
    if isinstance(column, np.ndarray):
        return column[i1_bytes:i2_bytes], offset
    else:
        return column.trim(i1_bytes, i2_bytes), offset

class ColumnStringArrow(ColumnString):
    """Column that unpacks the arrow string column on the fly"""
    def __init__(self, indices, bytes, length=None, offset=0, string_sequence=None, null_bitmap=None):
        self._string_sequence = string_sequence
        self.indices = indices
        self.offset = offset  # to avoid memory copies in trim
        self.bytes = bytes
        self.length = length if length is not None else len(indices) - 1
        if indices.dtype.kind == 'i' and indices.dtype.itemsize == 8:
            self.dtype = pa.large_string()
        elif indices.dtype.kind == 'i' and indices.dtype.itemsize == 4:
            self.dtype = pa.string()
        else:
            raise ValueError('unsupported index type' + str(indices.dtype))
        self.shape = (self.__len__(),)
        self.nbytes = self.bytes.nbytes + self.indices.nbytes
        self.null_bitmap = null_bitmap

        if not (self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize in [4,8]):
            raise ValueError('unsupported index type' + str(self.indices.dtype))

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.indices)
        hash2 = vaex.dataset.hash_array_data(self.bytes)
        fp = vaex.cache.fingerprint(hash1, hash2, self.length, self.offset, self.null_bitmap)
        return 'column-string-arrow-{fp}'

    def __arrow_array__(self, type=None):
        offsets = self.indices
        type = type or self.dtype
        if type == pa.string() and self.dtype == pa.large_string():
            offsets = offsets.astype(np.int32)  # downcast
        elif type == pa.large_string() and self.dtype == pa.string():
            type = pa.string()  # upcast

        # this code is very similar to vaex.arrow.convert.trim_buffers
        new_offset = self.offset % 8
        remove_offset = (self.offset // 8) * 8
        offsets = _asnumpy(_trim(offsets, remove_offset, self.offset + self.length + 1))
        first_offset = offsets[0]
        last_offset = offsets[new_offset + self.length]
        new_offsets = offsets - first_offset
        if self.null_bitmap is not None:
            null_bitmap, null_offset = _trim_bits(self.null_bitmap, self.offset, self.offset+self.length)
            assert null_offset == new_offset
            null_bitmap = pa.py_buffer(_asnumpy(null_bitmap))
        else:
            null_bitmap = None
        new_offsets_buffer = pa.py_buffer(new_offsets)
        bytes = pa.py_buffer(_asnumpy(_trim(self.bytes, first_offset, last_offset)))
        return pa.Array.from_buffers(type, self.length, [null_bitmap, new_offsets_buffer, bytes], offset=new_offset)

    @property
    def string_sequence(self):
        if self._string_sequence is None:
            if self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize == 8:
                string_type = vaex.strings.StringList64
            elif self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize == 4:
                string_type = vaex.strings.StringList32
            else:
                raise ValueError('unsupported index type' + str(self.indices.dtype))
            if self.offset == 0 and self.length == len(self.indices) - 1:
                # if the string is 'empty' with no indices
                # we just copy
                bytes = _asnumpy(self.bytes)
                indices = _asnumpy(self.indices)
                if self.null_bitmap is not None:
                    self._string_sequence = string_type(bytes, indices, self.length, self.offset, _asnumpy(self.null_bitmap), 0)
                else:
                    self._string_sequence = string_type(bytes, indices, self.length, self.offset)
            else:
                indices = _asnumpy(_trim(self.indices, self.offset, self.offset + self.length+1))
                if self.length:
                    byte_offset = indices[0]
                    byte_end = indices[self.length]
                    indices = indices - indices[0]
                    bytes = _asnumpy(_trim(self.bytes, byte_offset, byte_end))
                else:
                    bytes = _asnumpy(_trim(self.bytes, 0, 0))
                if self.null_bitmap is not None:
                    null_bitmap, null_offset = _trim_bits(self.null_bitmap, self.offset, self.offset+self.length)
                    self._string_sequence = string_type(bytes, indices, self.length, 0, _asnumpy(null_bitmap), null_offset)
                else:
                    self._string_sequence = string_type(bytes, indices, self.length, 0)
        return self._string_sequence

    def __len__(self):
        return self.length

    # TODO: in the future we might want to rely on these methods instead of special branches
    # in Expression for == and +. However, that gives different behaviour for dtype=object
    # (for instance you cannot use + when there is a None in it)
    def __eq__(self, other):
        if not isinstance(other, six.string_types):
            other = _to_string_sequence(other)
        return self.string_sequence.equals(other)

    def __add__(self, rhs):
        if not isinstance(rhs, six.string_types):
            rhs = _to_string_sequence(rhs)
        return self.from_string_sequence(self.string_sequence.concat(rhs))

    def __radd__(self, lhs):
        return self.from_string_sequence(self.string_sequence.concat_reverse(lhs))

    def __getitem__(self, slice):
        if isinstance(slice, int):
            return self.string_sequence.get(slice)
        elif isinstance(slice, np.ndarray):
            if np.ma.isMaskedArray(slice):
                if slice.dtype == np.bool_:
                    ss = self.string_sequence.index(slice.data & ~slice.mask)
                else:
                    ss = self.string_sequence.index(slice.data, slice.mask)
            else:
                ss = self.string_sequence.index(slice)
            return type(self).from_string_sequence(ss)
        else:
            start, stop, step = slice.start, slice.stop, slice.step
            start = start or 0
            stop = stop or len(self)
            assert step in [None, 1]
            return self.trim(start, stop)

    def to_numpy(self):
        return self.string_sequence.to_numpy()

    def trim(self, i1, i2):
        assert i2 >= i1
        return type(self)(self.indices, self.bytes, i2-i1, self.offset + i1, null_bitmap=self.null_bitmap)

    @classmethod
    def from_string_sequence(cls, string_sequence):
        s = string_sequence
        null_bitmap = s.null_bitmap
        if s.null_offset != 0:
            mask = ~s.mask()
            null_bitmap = np.frombuffer(pa.array(mask, pa.bool_()).buffers()[1], dtype='b')
        return cls(s.indices, s.bytes, s.length, s.offset, string_sequence=s, null_bitmap=null_bitmap)

    @classmethod
    def from_arrow(cls, ar):
        import vaex.arrow.convert
        return vaex.arrow.convert.column_from_arrow_array(ar)

    def _zeros_like(self):
        return ColumnStringArrow(np.zeros_like(self.indices), np.zeros_like(self.bytes), self.length, null_bitmap=self.null_bitmap)

    def get_mask(self):
        return self.string_sequence.mask()

    def astype(self, type):
        return self.to_numpy().astype(type)


from .s3fs import translate_options

def test_translate():
    assert translate_options({
        'anonymous': False,
        'access_key': 'acces',
        'secret_key': 'secret'
    }) == {
        'anon': False,
        'client_kwargs': {
            'aws_access_key_id': 'acces',
            'aws_secret_access_key': 'secret'
        }
    }


import vaex
import numpy as np
from .utils import _ensure_strings_from_expressions, _ensure_string_from_expression, as_flat_array
from .dataframe import docsubst


class DataFrameAccessorGeo(object):
    """Geometry/geographic helper methods

    Example:

    >>> df_xyz = df.geo.spherical2cartesian(df.longitude, df.latitude, df.distance)
    >>> df_xyz.x.mean()

    """
    def __init__(self, df):
        self.df = df

    @docsubst
    def spherical2cartesian(self, alpha, delta, distance, xname="x", yname="y", zname="z",
                            propagate_uncertainties=False,
                            center=[0, 0, 0], radians=False, inplace=False):
        """Convert spherical to cartesian coordinates.

        :param alpha:
        :param delta: polar angle, ranging from the -90 (south pole) to 90 (north pole)
        :param distance: radial distance, determines the units of x, y and z
        :param xname:
        :param yname:
        :param zname:
        :param propagate_uncertainties: {propagate_uncertainties}
        :param center:
        :param radians:
        :return: New dataframe (in inplace is False) with new x,y,z columns
        """
        df = self.df if inplace else self.df.copy()
        alpha = df._expr(alpha)
        delta = df._expr(delta)
        distance = df._expr(distance)
        if not radians:
            df.add_variable('pi', np.pi)
            alpha = alpha * df._expr('pi')/180
            delta = delta * df._expr('pi')/180

        # TODO: use sth like .optimize by default to get rid of the +0 ?
        if center[0]:
            df[xname] = np.cos(alpha) * np.cos(delta) * distance + center[0]
        else:
            df[xname] = np.cos(alpha) * np.cos(delta) * distance
        if center[1]:
            df[yname] = np.sin(alpha) * np.cos(delta) * distance + center[1]
        else:
            df[yname] = np.sin(alpha) * np.cos(delta) * distance
        if center[2]:
            df[zname] = np.sin(delta) * distance + center[2]
        else:
            df[zname] = np.sin(delta) * distance
        if propagate_uncertainties:
            df.propagate_uncertainties([df[xname], df[yname], df[zname]])
        return df

    def cartesian2spherical(self, x="x", y="y", z="z", alpha="l", delta="b", distance="distance", radians=False, center=None, center_name="solar_position", inplace=False):
        """Convert cartesian to spherical coordinates.



        :param x:
        :param y:
        :param z:
        :param alpha:
        :param delta: name for polar angle, ranges from -90 to 90 (or -pi to pi when radians is True).
        :param distance:
        :param radians:
        :param center:
        :param center_name:
        :return:
        """
        df = self.df if inplace else self.df.copy()
        df.add_variable('pi', np.pi)
        transform = "" if radians else "*180./pi"

        if center is not None:
            df.add_variable(center_name, center)
        if center is not None and center[0] != 0:
            x = "({x} - {center_name}[0])".format(**locals())
        if center is not None and center[1] != 0:
            y = "({y} - {center_name}[1])".format(**locals())
        if center is not None and center[2] != 0:
            z = "({z} - {center_name}[2])".format(**locals())
        df.add_virtual_column(distance, "sqrt({x}**2 + {y}**2 + {z}**2)".format(**locals()))
        # df.add_virtual_column(alpha, "((arctan2({y}, {x}) + 2*pi) % (2*pi)){transform}".format(**locals()))
        df.add_virtual_column(alpha, "arctan2({y}, {x}){transform}".format(**locals()))
        df.add_virtual_column(delta, "(-arccos({z}/{distance})+pi/2){transform}".format(**locals()))
        return df
    
    def cartesian_to_polar(self, x="x", y="y", radius_out="r_polar", azimuth_out="phi_polar",
                                               propagate_uncertainties=False,
                                               radians=False, inplace=False):
        """Convert cartesian to polar coordinates

        :param x: expression for x
        :param y: expression for y
        :param radius_out: name for the virtual column for the radius
        :param azimuth_out: name for the virtual column for the azimuth angle
        :param propagate_uncertainties: {propagate_uncertainties}
        :param radians: if True, azimuth is in radians, defaults to degrees
        :return:
        """
        df = self.df if inplace else self.df.copy()
        x = df._expr(x)
        y = df._expr(y)
        if radians:
            to_degrees = ""
        else:
            df.add_variable('pi', np.pi)
            to_degrees = "*180/pi"
        r = np.sqrt(x**2 + y**2)
        df[radius_out] = r
        phi = np.arctan2(y, x)
        if not radians:
            phi = phi * 180/np.pi
        df[azimuth_out] = phi
        if propagate_uncertainties:
            df.propagate_uncertainties([df[radius_out], df[azimuth_out]])
        return df

    @docsubst
    def velocity_polar2cartesian(self, x='x', y='y', azimuth=None, vr='vr_polar', vazimuth='vphi_polar', vx_out='vx', vy_out='vy', propagate_uncertainties=False, inplace=False):
        """ Convert cylindrical polar velocities to Cartesian.

        :param x:
        :param y:
        :param azimuth: Optional expression for the azimuth in degrees , may lead to a better performance when given.
        :param vr:
        :param vazimuth:
        :param vx_out:
        :param vy_out:
        :param propagate_uncertainties: {propagate_uncertainties}
        """
        df = self.df if inplace else self.df.copy()
        x = df._expr(x)
        y = df._expr(y)
        vr = df._expr(vr)
        vazimuth = df._expr(vazimuth)
        if azimuth is not None:
            azimuth = df._expr(azimuth)
            azimuth = np.deg2rad(azimuth)
        else:
            azimuth = np.arctan2(y, x)
        azimuth = df._expr(azimuth)
        df[vx_out] = vr * np.cos(azimuth) - vazimuth * np.sin(azimuth)
        df[vy_out] = vr * np.sin(azimuth) + vazimuth * np.cos(azimuth)
        if propagate_uncertainties:
            df.propagate_uncertainties([df[vx_out], df[vy_out]])
        return df

    @docsubst
    def velocity_cartesian2polar(self, x="x", y="y", vx="vx", radius_polar=None, vy="vy", vr_out="vr_polar", vazimuth_out="vphi_polar",
                                                          propagate_uncertainties=False, inplace=False):
        """Convert cartesian to polar velocities.

        :param x:
        :param y:
        :param vx:
        :param radius_polar: Optional expression for the radius, may lead to a better performance when given.
        :param vy:
        :param vr_out:
        :param vazimuth_out:
        :param propagate_uncertainties: {propagate_uncertainties}
        :return:
        """
        df = self.df if inplace else self.df.copy()
        x = df._expr(x)
        y = df._expr(y)
        vx = df._expr(vx)
        vy = df._expr(vy)
        if radius_polar is None:
            radius_polar = np.sqrt(x**2 + y**2)
        radius_polar = df._expr(radius_polar)
        df[vr_out]       = (x*vx + y*vy) / radius_polar
        df[vazimuth_out] = (x*vy - y*vx) / radius_polar
        if propagate_uncertainties:
            df.propagate_uncertainties([df[vr_out], df[vazimuth_out]])
        return df

    def velocity_cartesian2spherical(self, x="x", y="y", z="z", vx="vx", vy="vy", vz="vz", vr="vr", vlong="vlong", vlat="vlat", distance=None, inplace=False):
        """Convert velocities from a cartesian to a spherical coordinate system

        TODO: uncertainty propagation

        :param x: name of x column (input)
        :param y:         y
        :param z:         z
        :param vx:       vx
        :param vy:       vy
        :param vz:       vz
        :param vr: name of the column for the radial velocity in the r direction (output)
        :param vlong: name of the column for the velocity component in the longitude direction  (output)
        :param vlat: name of the column for the velocity component in the latitude direction, positive points to the north pole (output)
        :param distance: Expression for distance, if not given defaults to sqrt(x**2+y**2+z**2), but if this column already exists, passing this expression may lead to a better performance
        :return:
        """
        # see http://www.astrosurf.com/jephem/library/li110spherCart_en.htm
        df = self.df if inplace else self.df.copy()
        if distance is None:
            distance = "sqrt({x}**2+{y}**2+{z}**2)".format(**locals())
        df.add_virtual_column(vr, "({x}*{vx}+{y}*{vy}+{z}*{vz})/{distance}".format(**locals()))
        df.add_virtual_column(vlong, "-({vx}*{y}-{x}*{vy})/sqrt({x}**2+{y}**2)".format(**locals()))
        df.add_virtual_column(vlat, "-({z}*({x}*{vx}+{y}*{vy}) - ({x}**2+{y}**2)*{vz})/( {distance}*sqrt({x}**2+{y}**2) )".format(**locals()))
        return df

    def project_aitoff(self, alpha, delta, x, y, radians=True, inplace=False):
        """Add aitoff (https://en.wikipedia.org/wiki/Aitoff_projection) projection

        :param alpha: azimuth angle
        :param delta: polar angle
        :param x: output name for x coordinate
        :param y: output name for y coordinate
        :param radians: input and output in radians (True), or degrees (False)
        :return:
        """
        df = self.df if inplace else self.df.copy()
        transform = "" if radians else "*pi/180."
        aitoff_alpha = "__aitoff_alpha_%s_%s" % (alpha, delta)
        # sanatize
        aitoff_alpha = re.sub("[^a-zA-Z_]", "_", aitoff_alpha)

        df.add_virtual_column(aitoff_alpha, "arccos(cos({delta}{transform})*cos({alpha}{transform}/2))".format(**locals()))
        df.add_virtual_column(x, "2*cos({delta}{transform})*sin({alpha}{transform}/2)/sinc({aitoff_alpha}/pi)/pi".format(**locals()))
        df.add_virtual_column(y, "sin({delta}{transform})/sinc({aitoff_alpha}/pi)/pi".format(**locals()))
        return df

    def project_gnomic(self, alpha, delta, alpha0=0, delta0=0, x="x", y="y", radians=False, postfix="", inplace=False):
        """Adds a gnomic projection to the DataFrame"""
        df = self.df if inplace else self.df.copy()
        df.add_variable('pi', np.pi)
        if not radians:
            alpha = "pi/180.*%s" % alpha
            delta = "pi/180.*%s" % delta
            alpha0 = alpha0 * np.pi / 180
            delta0 = delta0 * np.pi / 180
        transform = "" if radians else "*180./pi"
        # aliases
        ra = alpha
        dec = delta
        ra_center = alpha0
        dec_center = delta0
        gnomic_denominator = 'sin({dec_center}) * tan({dec}) + cos({dec_center}) * cos({ra} - {ra_center})'.format(**locals())
        denominator_name = 'gnomic_denominator' + postfix
        xi = 'sin({ra} - {ra_center})/{denominator_name}{transform}'.format(**locals())
        eta = '(cos({dec_center}) * tan({dec}) - sin({dec_center}) * cos({ra} - {ra_center}))/{denominator_name}{transform}'.format(**locals())
        df.add_virtual_column(denominator_name, gnomic_denominator)
        df.add_virtual_column(x, xi)
        df.add_virtual_column(y, eta)
        return df

    def rotation_2d(self, x, y, xnew, ynew, angle_degrees, propagate_uncertainties=False, inplace=False):
        """Rotation in 2d.

        :param str x: Name/expression of x column
        :param str y: idem for y
        :param str xnew: name of transformed x column
        :param str ynew:
        :param float angle_degrees: rotation in degrees, anti clockwise
        :return:
        """
        df = self.df if inplace else self.df.copy()
        x = _ensure_string_from_expression(x)
        y = _ensure_string_from_expression(y)
        theta = np.radians(angle_degrees)
        matrix = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
        m = matrix_name = x + "_" + y + "_rot"
        for i in range(2):
            for j in range(2):
                df.set_variable(matrix_name + "_%d%d" % (i, j), matrix[i, j].item())
        df[xnew] = df._expr("{m}_00 * {x} + {m}_01 * {y}".format(**locals()))
        df[ynew] = df._expr("{m}_10 * {x} + {m}_11 * {y}".format(**locals()))
        if propagate_uncertainties:
            df.propagate_uncertainties([df[xnew], df[ynew]])
        return df

    def bearing(self, lon1, lat1, lon2, lat2, bearing="bearing", inplace=False):
        """Calculates a bearing, based on http://www.movable-type.co.uk/scripts/latlong.html"""
        df = self.df if inplace else self.df.copy()
        df.add_variable('pi', np.pi)
        lon1 = "(pickup_longitude * pi / 180)"
        lon2 = "(dropoff_longitude * pi / 180)"
        lat1 = "(pickup_latitude * pi / 180)"
        lat2 = "(dropoff_latitude * pi / 180)"
        p1 = lat1
        p2 = lat2
        l1 = lon1
        l2 = lon2
        # from http://www.movable-type.co.uk/scripts/latlong.html
        expr = "arctan2(sin({l2}-{l1}) * cos({p2}), cos({p1})*sin({p2}) - sin({p1})*cos({p2})*cos({l2}-{l1}))" \
            .format(**locals())
        df.add_virtual_column(bearing, expr)
        return df

    @docsubst
    def inside_which_polygons(self, x, y, pxss, pyss=None, any=True):
        """Find in which set of polygons (0 based index) a point resides.

        If any=True, it will be the first matching polygon set index, if any=False, it will
        be the first index that matches all polygons in the set.

        >>> import vaex
        >>> import numpy as np
        >>> df = vaex.from_arrays(x=[1, 2, 3], y=[2, 3, 4])
        >>> px = np.array([1.5, 2.5, 2.5, 1.5])
        >>> py = np.array([2.5, 2.5, 3.5, 3.5])
        >>> polygonA = [px, py]
        >>> polygonB = [px + 1, py + 1]
        >>> pxs = [[polygonA, polygonB], [polygonA]]
        >>> df['polygon_index'] = df.geo.inside_which_polygons(df.x, df.y, pxs, any=True)
        >>> df
        #    x    y  polygon_index
        0    1    2  --
        1    2    3  0
        2    3    4  0
        >>> df['polygon_index'] = df.geo.inside_which_polygons(df.x, df.y, pxs, any=False)
        >>> df
        #    x    y  polygon_index
        0    1    2  --
        1    2    3  1
        2    3    4  --

        :param x: {expression_one}
        :param y: {expression_one}
        :param px: list of N ndarrays with x coordinates for the polygon, N is the number of polygons
        :param px: list of N ndarrays with y coordinates for the polygon, if None, the shape of the ndarrays
                of the last dimention of the x arrays should be 2 (i.e. have the x and y coordinates)
        :param any: test if point it in any polygon (logically or), or all polygons (logically and)
        :return: Expression, 0 based index to which polygon the point belongs (or missing/masked value)
        """
        if pyss is None:
            list_of_polygons = pxss
            # polygon is a list of vertices shape: (N, 2)
            list_of_polygons = [[as_flat_array(np.asarray(polygon), np.float64) for polygon in polygons]
                                    for polygons in list_of_polygons]
            pxss = [[as_flat_array(polygon[0]) for polygon in polygons] for polygons in list_of_polygons]
            pyss = [[as_flat_array(polygon[1]) for polygon in polygons] for polygons in list_of_polygons]
        else:
            pxss = [[as_flat_array(np.asarray(pxs), np.float64) for pxs in polygons_x]
                                    for polygons_x in pxss]
            pyss = [[as_flat_array(np.asarray(pys), np.float64) for pys in polygons_y]
                                    for polygons_y in pyss]
        N = len(pxss)

        # iterate over polygon sets
        meanxss = []
        meanyss = []
        radiii = []
        for region_index in range(N):
            pxs = pxss[region_index]
            pys = pyss[region_index]
            meanxs = []
            meanys = []
            radii = []

            for polygons_index in range(len(pxs)):
                meanx = pxs[polygons_index].mean()
                meany = pys[polygons_index].mean()
                radius = np.sqrt((meanx - pxs[polygons_index])**2 + (meany - pys[polygons_index])**2).max()
                meanxs.append(meanx)
                meanys.append(meany)
                radii.append(radius)
            meanxss.append(meanxs)
            meanyss.append(meanys)
            radiii.append(radii)
        # this method is simply a wrapper around the true function, which also wants to have the
        # radii, which are precalculated for performance reasons
        return self.df.func.geo_inside_which_polygons(x, y, pxss, pyss, meanxss, meanyss, radiii, any=any)


@vaex.register_function(df_accessor=DataFrameAccessorGeo, name='inside_polygon')
def geo_inside_polygon(x, y, px, py):
    """Test if points defined by x and y are inside the polygon px, py

    Example:

    >>> import vaex
    >>> import numpy as np
    >>> df = vaex.from_arrays(x=[1, 2, 3], y=[2, 3, 4])
    >>> px = np.array([1.5, 2.5, 2.5, 1.5])
    >>> py = np.array([2.5, 2.5, 3.5, 3.5])
    >>> df['inside'] = df.geo.inside_polygon(df.x, df.y, px, py)
    >>> df
    #    x    y  inside
    0    1    2  False
    1    2    3  True
    2    3    4  False

    :param x: {expression_one}
    :param y: {expression_one}
    :param px: list of x coordinates for the polygon
    :param px: list of y coordinates for the polygon
    :return: Expression, which is true if point is inside, else false.
    """
    x = vaex.array_types.to_numpy(x)
    y = vaex.array_types.to_numpy(y)
    x = as_flat_array(x, np.float64)
    y = as_flat_array(y, np.float64)
    px = as_flat_array(np.asarray(px), np.float64)
    py = as_flat_array(np.asarray(py), np.float64)
    mask = np.zeros(len(x), dtype=bool)
    meanx = px.mean()
    meany = py.mean()
    radius = np.sqrt((meanx - px)**2 + (meany - py)**2).max()
    vaex.vaexfast.pnpoly(px, py, x, y, mask, meanx, meany, radius)
    return mask


@docsubst
@vaex.register_function(df_accessor=DataFrameAccessorGeo, name='inside_polygons')
def geo_inside_polygons(x, y, pxs, pys, any=True):
    """Test if points defined by x and y are inside all or any of the the polygons px, py

    Example:

    >>> import vaex
    >>> import numpy as np
    >>> df = vaex.from_arrays(x=[1, 2, 3], y=[2, 3, 4])
    >>> px = np.array([1.5, 2.5, 2.5, 1.5])
    >>> py = np.array([2.5, 2.5, 3.5, 3.5])
    >>> df['inside'] = df.geo.inside_polygons(df.x, df.y, [px, px + 1], [py, py + 1], any=True)
    >>> df
    #    x    y  inside
    0    1    2  False
    1    2    3  True
    2    3    4  True

    :param x: {expression_one}
    :param y: {expression_one}
    :param pxs: list of N ndarrays with x coordinates for the polygon, N is the number of polygons
    :param pxs: list of N ndarrays with y coordinates for the polygon
    :param any: return true if in any polygon, or all polygons
    :return: Expression , which is true if point is inside, else false.
    """
    x = vaex.array_types.to_numpy(x)
    y = vaex.array_types.to_numpy(y)
    x = as_flat_array(x, np.float64)
    y = as_flat_array(y, np.float64)
    mask = np.zeros(len(x), dtype=bool)

    N = len(pxs)
    submask = np.zeros(len(x), dtype=bool)
    if N > 0:
        pxs = [vaex.arrow.numpy_dispatch.unwrap(k) for k in pxs]
        pys = [vaex.arrow.numpy_dispatch.unwrap(k) for k in pys]
        for i in range(0, N):
            px = as_flat_array(pxs[i], np.float64)
            py = as_flat_array(pys[i], np.float64)
            meanx = px.mean()
            meany = py.mean()
            radius = np.sqrt((meanx - px)**2 + (meany - py)**2).max()
            vaex.vaexfast.pnpoly(px, py, x, y, submask if i > 0 else mask, meanx, meany, radius)
            if i > 0:
                if any:
                    mask = mask | submask
                else:
                    mask = mask & submask
    return mask


@docsubst
@vaex.register_function(df_accessor=DataFrameAccessorGeo, name='inside_which_polygon')
def geo_inside_which_polygon(x, y, pxs, pys):
    """Find in which polygon (0 based index) a point resides

    Example:

    >>> import vaex
    >>> import numpy as np
    >>> df = vaex.from_arrays(x=[1, 2, 3], y=[2, 3, 4])
    >>> px = np.array([1.5, 2.5, 2.5, 1.5])
    >>> py = np.array([2.5, 2.5, 3.5, 3.5])
    >>> df['polygon_index'] = df.geo.inside_which_polygon(df.x, df.y, [px, px + 1], [py, py + 1])
    >>> df
    #    x    y  polygon_index
    0    1    2  --
    1    2    3  0
    2    3    4  1

    :param x: {expression_one}
    :param y: {expression_one}
    :param px: list of N ndarrays with x coordinates for the polygon, N is the number of polygons
    :param px: list of N ndarrays with y coordinates for the polygon
    :return: Expression, 0 based index to which polygon the point belongs (or missing/masked value)
    """
    x = vaex.array_types.to_numpy(x)
    y = vaex.array_types.to_numpy(y)
    x = as_flat_array(x, np.float64)
    y = as_flat_array(y, np.float64)
    # list do not get unwrapped
    pxs = [vaex.arrow.numpy_dispatch.unwrap(k) for k in pxs]
    pys = [vaex.arrow.numpy_dispatch.unwrap(k) for k in pys]
    pxs = as_flat_array(np.asarray(pxs), np.float64)
    pys = as_flat_array(np.asarray(pys), np.float64)
    polygon_indices = np.zeros(len(x), dtype=np.int32)
    polygon_mask = np.ones(len(x), dtype=bool)
    polygon_indices = np.ma.array(polygon_indices, mask=polygon_mask)
    indices = np.arange(len(x), dtype=np.uint32)
    inside_mask = np.zeros(len(x), dtype=bool)

    meanx = pxs.mean()
    meany = pys.mean()
    radius = np.sqrt((meanx - pxs)**2 + (meany - pys)**2).max()
    N = len(pxs)
    # mask = mask | submask
    for i in range(N):
        vaex.vaexfast.pnpoly(pxs[i], pys[i], x, y, inside_mask, meanx, meany, radius)
        # mark all points that are inside with the index of the polygon
        polygon_indices.data[indices[inside_mask]] = i
        polygon_indices.mask[indices[inside_mask]] = False
        # now remove all values that found a matching polygon
        x = x[~inside_mask]
        y = y[~inside_mask]
        # and keep track of where they point to in the polygon_indices array
        indices = indices[~inside_mask]
        # trim mask
        inside_mask = inside_mask[:len(x)]
        if len(x) == 0:
            break
    return polygon_indices


@docsubst
@vaex.register_function()
def geo_inside_which_polygons(x, y, pxss, pyss, meanxss, meanyss, radiii, any):
    # real implementation of geo.inside_which_polygon
    x = vaex.array_types.to_numpy(x)
    y = vaex.array_types.to_numpy(y)
    x = as_flat_array(x, np.float64)
    y = as_flat_array(y, np.float64)
    polygon_indices = np.zeros(len(x), dtype=np.int32)
    polygon_mask = np.ones(len(x), dtype=bool)
    polygon_indices = np.ma.array(polygon_indices, mask=polygon_mask)
    indices = np.arange(len(x), dtype=np.uint32)
    inside_mask = np.zeros(len(x), dtype=bool)
    N = len(pxss)
    for i in range(N):
        pxs = pxss[i]
        pys = pyss[i]
        pxs = [vaex.arrow.numpy_dispatch.unwrap(k) for k in pxs]
        pys = [vaex.arrow.numpy_dispatch.unwrap(k) for k in pys]
        meanxs = meanxss[i]
        meanys = meanyss[i]
        radii = radiii[i]
        M = len(pxs)
        if M > 0:
            vaex.vaexfast.pnpoly(pxs[0], pys[0], x, y, inside_mask, meanxs[0], meanys[0], radii[0])
        if M > 1:
            inside_sub_mask = np.zeros(len(x), dtype=bool)
            for j in range(1, M):
                vaex.vaexfast.pnpoly(pxs[j], pys[j], x, y, inside_sub_mask, meanxs[j], meanys[j], radii[j])
                if any:
                    inside_mask = inside_mask | inside_sub_mask
                else:
                    inside_mask = inside_mask & inside_sub_mask

        # mark all points that are inside with the index of the polygon
        polygon_indices.data[indices[inside_mask]] = i
        polygon_indices.mask[indices[inside_mask]] = False
        # no remove all values that found a matching polygon
        x = x[~inside_mask]
        y = y[~inside_mask]
        # and keep track of where they point to in the polygon_indices array
        indices = indices[~inside_mask]
        # trim mask
        inside_mask = inside_mask[:len(x)]
        if len(x) == 0:
            break
    return polygon_indices



import difflib
import logging
import os
import operator
from functools import reduce
import threading
import pkg_resources

import numpy as np

import vaex.promise
import vaex.settings
import vaex.encoding
from .utils import _expand_shape
from .datatype import DataType


logger = logging.getLogger('vaex.tasks')
register = vaex.encoding.make_class_registery('task')

_lock = threading.Lock()
_task_checker_types = {}


class Checker:
    track_live = False

    def add_task(self, task):
        pass

    def add_task_part(self, task):
        pass


def create_checkers():
    if not _task_checker_types:
        with _lock:
            if not _task_checker_types:
                for entry in pkg_resources.iter_entry_points(group="vaex.task.checker"):
                    _task_checker_types[entry.name] = entry.load()
    names = list(_task_checker_types.keys())
    task_checkers_type = vaex.settings.main.task_tracker.type
    types = [k for k in task_checkers_type.split(",") if k]
    checkers = []
    for type in types:
        cls = _task_checker_types.get(type)
        if cls is not None:
            checkers.append(cls())
        else:
            msg = f"Task checker {type} does not exist."
            matches = difflib.get_close_matches(type, names)
            if matches:
                msg += " Did you mean: " + " or ".join(map(repr, matches))
            else:
                msg += " Options are: " + repr(names)
            raise NameError(msg)
    return checkers


class TaskCheckError(RuntimeError):
    pass


class Task(vaex.promise.Promise):
    """
    :type: signal_progress: Signal
    """
    _fingerprint = None
    requires_fingerprint = False
    cacheable = True
    see_all = False
    _toreject: Exception = None

    def __init__(self, df=None, expressions=[], pre_filter=False, name="task"):
        vaex.promise.Promise.__init__(self)
        self.df = df
        self.expressions = expressions
        self.expressions_all = list(expressions)
        self.signal_start = vaex.events.Signal("start of task")
        self.signal_progress = vaex.events.Signal("progress (float)")
        self.progress_fraction = 0
        self.signal_progress.connect(self._set_progress)
        self.cancelled = False
        self.stopped = False  # a task can stop early, but without error
        self.name = name
        self.pre_filter = pre_filter
        self.result = None

    def dependencies(self):
        variables = set()
        for expression in self.expressions_all:
            variables |= self.df._expr(expression).dependencies()
        for expression in self.selections:
            variables |= self.df._selection_expression(expression).dependencies()
        return variables

    def fingerprint(self):
        if self._fingerprint is None:
            dependencies = self.dependencies()
            df_fp = self.df.fingerprint(dependencies=dependencies)
            task_fp = vaex.encoding.fingerprint('task', self)
            self._fingerprint = f'task-{self.name}-{task_fp}-{df_fp}'
        return self._fingerprint

    def progress(self, fraction):
        if not self.cancelled:
            self.cancelled = not all(self.signal_progress.emit(fraction))
        return not self.cancelled

    def _set_progress(self, fraction):
        self.progress_fraction = fraction
        return not self.cancelled  # don't cancel

    def cancel(self):
        self.cancelled = True

    @property
    def dimension(self):
        return len(self.expressions)

    @classmethod
    def create(cls):
        ret = Task()
        return ret

    def create_next(self):
        ret = Task(self.df, [])
        self.signal_progress.connect(lambda f: all(ret.signal_progress.emit(f)))
        return ret

    # def __repr__(self):
    #     from .encoding import Encoding
    #     encoding = Encoding()
    #     return json.dumps({key: repr(value) for key, value in self.encode(encoding).items()})


# only used for testing
@register
class TaskSum(Task):
    snake_name = "sum-test"

    def get_bin_count(self):
        return 1

    # def __init__(self, df, expression):
    #     super().__init__(df, expression)
    #     self.expression = expression

    def encode(self, encoding):
        return {'expression': self.expressions}

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(df, expression=[spec['expression']])


@register
class TaskFilterFill(Task):
    cacheable = False
    snake_name = "filter_fill"
    def __init__(self, df):
        super().__init__(df=df, pre_filter=True, name=self.snake_name)
        self.selections = []

    def get_bin_count(self):
        return 0

    def encode(self, encoding):
        return {}

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(df)

@register
class TaskHashmapUniqueCreate(Task):
    see_all = True
    requires_fingerprint = True
    snake_name = "hash_map_unique_create"
    def __init__(self, df, expression, flatten, limit=None, limit_raise=True, selection=None, return_inverse=False):
        super().__init__(df=df, expressions=[expression], pre_filter=df.filtered, name=self.snake_name)
        self.flatten = flatten
        self.dtype = self.df.data_type(expression)
        self.dtype_item = self.df.data_type(expression, axis=-1 if flatten else 0)
        self.limit = limit
        self.limit_raise = limit_raise
        self.selection = selection
        self.selections = [self.selection]
        self.return_inverse = return_inverse

    def get_bin_count(self):
        return 0

    def __repr__(self):
        return f"task-{self.snake_name}: expression={self.expressions[0]!r}"

    def encode(self, encoding):
        return {'expression': self.expressions[0], 'dtype': encoding.encode('dtype', self.dtype),
                'dtype_item': encoding.encode('dtype', self.dtype_item), 'flatten': self.flatten,
                'limit': self.limit, 'limit_raise': self.limit_raise,
                'selection': self.selection, 'return_inverse': self.return_inverse}

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(df, expression=spec['expression'])


class TaskValueCounts(Task):
    snake_name = "value_counts"

    def __init__(self, df, expression, dropnan=False, dropmissing=False, ascending=False, axis=None):
        super().__init__(df=df, expressions=[expression], pre_filter=df.filtered, name=self.snake_name)
        self.flatten = axis is None
        self.dtype = self.df.data_type(expression)
        self.dtype_item = self.df.data_type(expression, axis=-1 if axis is None else 0)
        self.selection = None
        self.selections = [self.selection]
        self.dropnan = dropnan
        self.dropmissing = dropmissing
        self.ascending = ascending

    def get_bin_count(self):
        return 0

    def __repr__(self):
        return f"task-{self.snake_name}: expression={self.expressions[0]!r}"

    def encode(self, encoding):
        return {
            "expression": self.expressions[0],
            "dtype": encoding.encode("dtype", self.dtype),
            "dtype_item": encoding.encode("dtype", self.dtype_item),
            "flatten": self.flatten,
            "dropnan": self.dropnan,
            "dropmissing": self.dropmissing,
            "ascending": self.ascending,
        }

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(df, expression=spec["expression"])


@register
class TaskMapReduce(Task):
    snake_name = "map_reduce"
    cacheable = False  # in general this is used with side effects

    def __init__(self, df, expressions, map, reduce, info=False, to_float=False,
                 to_numpy=True, skip_masked=False, ignore_filter=False, selection=None, pre_filter=False, name="task"):
        Task.__init__(self, df, expressions, name=name, pre_filter=pre_filter)
        self._map = map
        self._reduce = reduce
        self.info = info
        self.to_float = to_float
        self.to_numpy = to_numpy
        self.skip_masked = skip_masked
        self.ignore_filter = ignore_filter
        if self.pre_filter and self.ignore_filter:
            raise ValueError("Cannot pre filter and also ignore the filter")
        self.selection = selection
        self.selections = [self.selection]

    def get_bin_count(self):
        return 0

    def encode(self, encoding):
        return {'expressions': self.expressions, 'map': self._map, 'reduce': self._reduce,
                'info': self.info, 'to_float': self.to_float, 'to_numpy': self.to_numpy,  # 'ordered_reduce': self.ordered_reduce,
                'skip_masked': self.skip_masked, 'ignore_filter': self.ignore_filter, 'selection': self.selection, 'pre_filter': self.pre_filter,
                }

    def __repr__(self) -> str:
        return f'TaskMapReduce(map={self._map}, reduce={self._reduce}'


class StatOp(object):
    def __init__(self, code, fields, reduce_function=np.nansum, dtype=None):
        self.code = code
        self.fixed_fields = fields
        self.reduce_function = reduce_function
        self.dtype = dtype

    def init(self, grid):
        pass

    def fields(self, weights):
        return self.fixed_fields

    def reduce(self, grid, axis=0):
        value = self.reduce_function(grid, axis=axis)
        if self.dtype:
            return value.astype(self.dtype)
        else:
            return value

    def encode(self, encoding):
        return {'code': self.code, 'fields': self.fixed_fields, 'reduce_function': self.reduce_function.__name__}


class StatOpMinMax(StatOp):
    def __init__(self, code, fields):
        super(StatOpMinMax, self).__init__(code, fields)

    def init(self, grid):
        grid[..., 0] = np.inf
        grid[..., 1] = -np.inf

    def reduce(self, grid, axis=0):
        out = np.zeros(grid.shape[1:], dtype=grid.dtype)
        out[..., 0] = np.nanmin(grid[..., 0], axis=axis)
        out[..., 1] = np.nanmax(grid[..., 1], axis=axis)
        return out

    def encode(self, encoding):
        return {'code': self.code, 'fields': self.fixed_fields}


class StatOpCov(StatOp):
    def __init__(self, code, fields=None, reduce_function=np.sum):
        super(StatOpCov, self).__init__(code, fields, reduce_function=reduce_function)

    def fields(self, weights):
        N = len(weights)
        # counts, sums, cross product sums
        return N * 2 + N**2 * 2  # ((N+1) * N) // 2 *2

    def encode(self, encoding):
        return {'code': self.code, 'reduce_function': self.reduce_function.__name__}


class StatOpFirst(StatOp):
    def __init__(self, code):
        super(StatOpFirst, self).__init__(code, 2, reduce_function=self._reduce_function)

    def init(self, grid):
        grid[..., 0] = np.nan
        grid[..., 1] = np.inf

    def _reduce_function(self, grid, axis=0):
        values = grid[..., 0]
        order_values = grid[..., 1]
        indices = np.argmin(order_values, axis=0)

        # see e.g. https://stackoverflow.com/questions/46840848/numpy-how-to-use-argmax-results-to-get-the-actual-max?noredirect=1&lq=1
        # and https://jakevdp.github.io/PythonDataScienceHandbook/02.07-fancy-indexing.html
        if len(values.shape) == 2:  # no binby
            return values[indices, np.arange(values.shape[1])[:, None]][0]
        if len(values.shape) == 3:  # 1d binby
            return values[indices, np.arange(values.shape[1])[:, None], np.arange(values.shape[2])]
        if len(values.shape) == 4:  # 2d binby
            return values[indices, np.arange(values.shape[1])[:, None], np.arange(values.shape[2])[None, :, None], np.arange(values.shape[3])]
        else:
            raise ValueError('dimension %d not yet supported' % len(values.shape))

    def fields(self, weights):
        # the value found, and the value by which it is ordered
        return 2

    def encode(self, encoding):
        return {'code': self.code, 'reduce_function': self.reduce_function.__name__}


@vaex.encoding.register('_op')
class op_encoding:
    @staticmethod
    def encode(encoding, op):
        return op.encode(encoding)

    @staticmethod
    def decode(encoding, op_spec):
        op_spec = op_spec.copy()
        if 'reduce_function' in op_spec:
            op_spec['reduce_function'] = getattr(np, op_spec.pop('reduce_function'))
        cls = StatOp
        if op_spec['code'] == 2:
            cls = StatOpMinMax
        if op_spec['code'] == 5:
            cls = StatOpCov
        if op_spec['code'] == 6:
            cls = StatOpFirst
        return cls(**op_spec)


OP_ADD1 = StatOp(0, 1)
OP_COUNT = StatOp(1, 1)
OP_MIN_MAX = StatOpMinMax(2, 2)
OP_ADD_WEIGHT_MOMENTS_01 = StatOp(3, 2, np.nansum)
OP_ADD_WEIGHT_MOMENTS_012 = StatOp(4, 3, np.nansum)
OP_COV = StatOpCov(5)
OP_FIRST = StatOpFirst(6)


@register
class TaskStatistic(Task):
    snake_name = "legacy_statistic"

    def encode(self, encoding):
        return {'expressions': self.expressions,
                'shape': self.shape, 'selections': self.selections, 'op': encoding.encode('_op', self.op), 'weights': self.weights,
                'dtype': encoding.encode('dtype', DataType(self.dtype)), 'minima': self.minima, 'maxima': self.maxima, 'edges': self.edges,
                'selection_waslist': self.selection_waslist}

    def get_bin_count(self):
        return reduce(operator.mul, self.shape, 1)

    @classmethod
    def decode(cls, encoding, spec, df):
        spec = spec.copy()
        spec['op'] = encoding.decode('_op', spec['op'])
        spec['dtype'] = encoding.decode('dtype', spec['dtype'])
        selection_waslist = spec.pop('selection_waslist')
        if selection_waslist:
            spec['selection'] = spec.pop('selections')
        else:
            spec['selection'] = spec.pop('selections')[0]
        spec['limits'] = list(zip(spec.pop('minima'), spec.pop('maxima')))
        return cls(df, **spec)

    def __init__(self, df, expressions, shape, limits, masked=False, weights=[], weight=None, op=OP_ADD1, selection=None, edges=False,
                 dtype=np.dtype('f8')):
        if not isinstance(expressions, (tuple, list)):
            expressions = [expressions]
        # edges include everything outside at index 1 and -1, and nan's at index 0, so we add 3 to each dimension
        self.shape = tuple([k + 3 if edges else k for k in _expand_shape(shape, len(expressions))])
        self.limits = limits
        if weight is not None:  # shortcut for weights=[weight]
            assert weights == [], 'only provide weight or weights, not both'
            weights = [weight]
            del weight
        self.weights = weights
        self.selection_waslist, [self.selections, ] = vaex.utils.listify(selection)
        self.op = op
        self.edges = edges
        Task.__init__(self, df, expressions, name="statisticNd", pre_filter=df.filtered)
        # self.dtype = np.int64 if self.op == OP_ADD1 else np.float64 # TODO: use int64 fir count and ADD1
        self.dtype = dtype
        self.masked = masked

        self.fields = op.fields(weights)
        # self.shape_total = (self.df.executor.thread_pool.nthreads,) + (len(self.selections), ) + self.shape + (self.fields,)
        # self.grid = np.zeros(self.shape_total, dtype=self.dtype)
        # self.op.init(self.grid)
        self.minima = []
        self.maxima = []
        limits = np.array(self.limits)
        if len(limits) != 0:
            logger.debug("limits = %r", limits)
            assert limits.shape[-1] == 2, "expected last dimension of limits to have a length of 2 (not %d, total shape: %s), of the form [[xmin, xmin], ... [zmin, zmax]], not %s" %\
                                          (limits.shape[-1], limits.shape, limits)
            if len(limits.shape) == 1:  # short notation: [xmin, max], instead of [[xmin, xmax]]
                limits = [limits]
            logger.debug("limits = %r", limits)
            for limit in limits:
                vmin, vmax = limit
                self.minima.append(float(vmin))
                self.maxima.append(float(vmax))
        # if self.weight is not None:
        self.expressions_all.extend(weights)


@register
class TaskAggregations(Task):
    """Multiple aggregations on a single grid."""
    snake_name = "aggregations"
    see_all = True

    def __init__(self, df, binners):
        expressions = [binner.expression for binner in binners]
        self.df = df
        self.binners = binners
        self.aggregation_descriptions = []
        self.dtypes = {}
        Task.__init__(self, df, expressions, name="statisticNd", pre_filter=df.filtered)
        self.original_tasks = []
        self.selections = []

    def get_bin_count(self):
        return reduce(lambda prev, binner: binner.count * prev, self.binners, 1)

    def __repr__(self):
        encoding = vaex.encoding.Encoding()
        state = self.encode(encoding)
        import yaml
        return yaml.dump(state, sort_keys=False, indent=4)

    def encode(self, encoding):
        # TODO: get rid of dtypes
        return {
                'binners': encoding.encode_list('binner', self.binners),
                'aggregations': encoding.encode_list("aggregation", self.aggregation_descriptions),
                'dtypes': encoding.encode_dict("dtype", self.dtypes)
                }

    @classmethod
    def decode(cls, encoding, spec, df):
        binners = encoding.decode_list('binner', spec['binners'])
        task = cls(df, binners)
        aggs = encoding.decode_list('aggregation', spec['aggregations'])
        for agg in aggs:
            agg._prepare_types(df)
            task.add_aggregation_operation(agg)
        return task

    def add_aggregation_operation(self, aggregator_descriptor):
        assert self._fingerprint is None, "Adding operation after fingerprint is calculated"
        task = Task(self.df, [], "--")

        def chain_reject(x):
            task.reject(x)
            return x

        def assign_subtask(values, index=len(self.aggregation_descriptions)):
            task.fulfill(values[index])
        self.then(assign_subtask, chain_reject)

        self.aggregation_descriptions.append((aggregator_descriptor))
        if aggregator_descriptor.selection is not None:
            if isinstance(aggregator_descriptor.selection, (list, tuple)):
                self.selections.extend([str(selection) if selection is not None else None for selection in aggregator_descriptor.selection])
            else:
                self.selections.append(str(aggregator_descriptor.selection))
        else:
            self.selections.append(None)
        # THIS SHOULD BE IN THE SAME ORDER AS THE ABOVE TASKPART
        # it is up the the executor to remove duplicate expressions
        self.expressions_all.extend(aggregator_descriptor.expressions)
        # TODO: optimize/remove?
        self.dtypes = {expr: self.df.data_type(expr).index_type for expr in self.expressions_all}
        return task

    def check(self):
        if not self.aggregation_descriptions:
            raise RuntimeError('Aggregation tasks started but nothing to do, maybe adding operations failed?')


@register
class TaskAggregation(Task):
    """Single aggregation on a single grid."""
    snake_name = "aggregation-single"

    def __init__(self, df, binners, aggregation_description):
        expressions = [binner.expression for binner in binners]
        self.df = df
        self.binners = binners
        self.aggregation_description = aggregation_description
        self.dtypes = {}
        Task.__init__(self, df, expressions, name=self.snake_name, pre_filter=df.filtered)
        self.dtypes = {expr: self.df.data_type(expr).index_type for expr in self.expressions_all}
        self.expressions_all.extend(aggregation_description.expressions)
        self.selections = [str(aggregation_description.selection) if aggregation_description.selection is not None else None]

    def __repr__(self):
        return f"task-{self.snake_name} agg={self.aggregation_description!r} selection={self.selections[0]!r} binners=[{self.binners!r}]"

    def encode(self, encoding):
        # TODO: get rid of dtypes
        return {
                'binners': encoding.encode_list('binner', self.binners),
                'aggregation': encoding.encode("aggregation", self.aggregation_description),
                'dtypes': encoding.encode_dict("dtype", self.dtypes)
                }

    @classmethod
    def decode(cls, encoding, spec, df):
        binners = tuple(encoding.decode_list('binner', spec['binners']))
        agg = encoding.decode('aggregation', spec['aggregation'])
        task = cls(df, binners, agg)
        return task



"""Implementation all tasks parts for the cpu"""

from functools import reduce
import logging
import operator
import sys
from typing import List

import numpy as np
import pyarrow as pa

import vaex
import vaex.encoding
from vaex.hash import counter_type_from_dtype
from .utils import as_flat_float, as_flat_array, _issequence, _ensure_list
from .array_types import filter


logger = logging.getLogger("vaex.cpu")
register = vaex.encoding.make_class_registery('task-part-cpu')



@vaex.encoding.register("binner-cpu")
class binner_encoding:
    @staticmethod
    def encode(encoding, binner):
        raise RuntimeError('binner-cpu should not get encoded')
        # name = type(binner).__name__
        # if name.startswith('BinnerOrdinal_'):
        #     dtype = name[len('BinnerOrdinal_'):]
        #     if dtype.endswith("_non_native"):
        #         dtype = dtype[:-len('64_non_native')]
        #         dtype = encoding.encode('dtype', DataType(np.dtype(dtype).newbyteorder()))
        #     return {'type': 'ordinal', 'expression': binner.expression, 'dtype': dtype, 'count': binner.ordinal_count, 'minimum': binner.min_value}
        # elif name.startswith('BinnerScalar_'):
        #     dtype = name[len('BinnerScalar_'):]
        #     if dtype.endswith("_non_native"):
        #         dtype = dtype[:-len('64_non_native')]
        #         dtype = encoding.encode('dtype', DataType(np.dtype(dtype).newbyteorder()))
        #     return {'type': 'scalar', 'expression': binner.expression, 'dtype': dtype, 'count': binner.bins, 'minimum': binner.vmin, 'maximum': binner.vmax}

    @staticmethod
    def decode(encoding, binner_spec, nthreads):
        type = binner_spec['binner-type']
        dtype = encoding.decode('dtype', binner_spec['dtype'])
        if type == 'ordinal':
            cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerOrdinal_", dtype)
            return cls(nthreads, binner_spec["expression"], binner_spec["count"], binner_spec["minimum"], False, binner_spec["invert"])
        elif type == "scalar":
            cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerScalar_", dtype)
            return cls(nthreads, binner_spec["expression"], binner_spec["minimum"], binner_spec["maximum"], binner_spec["count"])
        elif type == "hash":
            cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerHash_", dtype)
            hash_map_unique_id = binner_spec["hash_map_unique"]
            if encoding.has_object(hash_map_unique_id):
                hash_map_unique = encoding.get_object(hash_map_unique_id)
            else:
                hash_map_unique_spec = encoding.get_object_spec(hash_map_unique_id)
                hash_map_unique = encoding.decode("hash-map-unique", hash_map_unique_spec)
                encoding.set_object(hash_map_unique_id, hash_map_unique)
            return cls(binner_spec["expression"], hash_map_unique._internal)
        else:
            raise ValueError('Cannot deserialize: %r' % binner_spec)

class TaskPart:
    stopped = False
    def __init__(self, df, expressions, name, pre_filter):
        self.df = df
        self.expressions = expressions
        self.name = name
        self.pre_filter = pre_filter

    def ideal_splits(self, nthreads):
        return nthreads

    def memory_usage(self):
        return 0


@register
class TaskPartSum(TaskPart):
    snake_name = "sum-test"

    def __init__(self, expression):
        self.total = 0
        self.expression = expression

    def get_bin_count(self):
        return 1

    @property
    def expressions(self):
        return [self.expression]

    def get_result(self):
        return self.total

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        self.total += blocks[0].sum()

    def reduce(self, others):
        self.total += sum(other.total for other in others)

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(spec['expression'])



@register
class TaskPartFilterFill(TaskPart):
    snake_name = "filter_fill"

    def __init__(self):
        pass

    def get_bin_count(self):
        return 0

    @property
    def expressions(self):
        return []

    def get_result(self):
        return None

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        assert filter_mask is not None, f'{filter_mask}'

    def reduce(self, others):
        pass

    @classmethod
    def decode(cls, encoding, spec, df, nthreads):
        assert spec == {}
        return cls()


@register
class TaskPartValueCounts(TaskPart):
    snake_name = "value_counts"

    def __init__(self, df, expression, dtype, dtype_item, dropnan, dropmissing, ascending, flatten, nthreads):
        super().__init__(df=df, expressions=[expression], pre_filter=df.filtered, name=self.snake_name)
        self.flatten = flatten
        self.dtype = dtype
        self.dtype_item = dtype_item
        self.data_type = vaex.dtype(self.dtype)
        self.data_type_item = vaex.dtype(self.dtype_item)
        self.selection = None
        self.selections = [self.selection]
        self.dropnan = dropnan
        self.dropmissing = dropmissing
        self.ascending = ascending
        transient = False
        # TODO: in the future we might want to explore this again
        # transient = self.transient or self.ds.filtered or self.ds.is_masked(self.expression)
        # if self.is_string() and not transient:
        #     # string is a special case, only ColumnString are not transient
        #     ar = self.ds.columns[self.expression]
        #     from vaex.column import ColumnString
        #     if not isinstance(ar, ColumnString):
        #         transient = True
        self.counter_type = counter_type_from_dtype(self.data_type_item, transient)
        self.counters = [None] * nthreads
        self.result = None

    @classmethod
    def decode(cls, encoding, spec, df, nthreads):
        return cls(
            df,
            spec["expression"],
            encoding.decode("dtype", spec["dtype"]),
            encoding.decode("dtype", spec["dtype_item"]),
            dropnan=spec["dropnan"],
            dropmissing=spec["dropmissing"],
            ascending=spec["ascending"],
            flatten=spec["flatten"],
            nthreads=nthreads,
        )

    def memory_usage(self):
        return sum((sys.getsizeof(k) for k in self.counters if k is not None), 0)

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        ar = blocks[0]
        if len(ar) == 0:
            return 0
        if self.counters[thread_index] is None:
            self.counters[thread_index] = self.counter_type(1)
        if self.data_type.is_list and self.flatten:
            try:
                ar = ar.values
            except AttributeError:  # pyarrow ChunkedArray
                ar = ar.combine_chunks().values
        if self.data_type_item.is_string:
            from vaex.column import _to_string_sequence

            ar = _to_string_sequence(ar)
        else:
            ar = vaex.array_types.to_numpy(ar)
        if np.ma.isMaskedArray(ar):
            mask = np.ma.getmaskarray(ar)
            self.counters[thread_index].update(ar, mask)
        else:
            self.counters[thread_index].update(ar)
        return 0

    def reduce(self, others: List["TaskPartValueCounts"]):
        counters = []
        for other in [self, *others]:
            counters.extend([k for k in other.counters if k is not None])
        counter = counters[0]
        for other in counters[1:]:
            counter.merge(other)
        if self.data_type_item.is_object:
            # for dtype=object we use the old interface
            # since we don't care about multithreading (cannot release the GIL)
            key_values = counter.extract()
            keys = list(key_values.keys())
            counts = list(key_values.values())
            if counter.has_nan and not self.dropnan:
                keys = [np.nan] + keys
                counts = [counter.nan_count] + counts
            if counter.has_null and not self.dropmissing:
                keys = [None] + keys
                counts = [counter.null_count] + counts
            if self.dropmissing and None in keys:
                # we still can have a None in the values
                index = keys.index(None)
                keys.pop(index)
                counts.pop(index)
            counts = np.array(counts)
            keys = np.array(keys)
        else:
            keys = counter.key_array()
            counts = counter.counts()
            if isinstance(keys, (vaex.strings.StringList32, vaex.strings.StringList64)):
                keys = vaex.strings.to_arrow(keys)

            deletes = []
            if counter.has_nan:
                null_offset = 1
            else:
                null_offset = 0
            if self.dropmissing and counter.has_null:
                deletes.append(counter.null_index)
            if self.dropnan and counter.has_nan:
                deletes.append(counter.nan_index)
            if vaex.array_types.is_arrow_array(keys):
                indices = np.delete(np.arange(len(keys)), deletes)
                keys = keys.take(indices)
            else:
                keys = np.delete(keys, deletes)
                if not self.dropmissing and counter.has_null:
                    mask = np.zeros(len(keys), dtype=np.uint8)
                    mask[null_offset] = 1
                    keys = np.ma.array(keys, mask=mask)
            counts = np.delete(counts, deletes)

        order = np.argsort(counts)
        if not self.ascending:
            order = order[::-1]
        counts = counts[order]
        keys = keys.take(order)

        keys = keys.tolist()
        if None in keys:
            index = keys.index(None)
            keys.pop(index)
            keys = ["missing"] + keys
            counts = counts.tolist()
            count_null = counts.pop(index)
            counts = [count_null] + counts

        from pandas import Series

        self.result = Series(counts, index=keys)

    def get_result(self):
        return self.result

@register
class TaskPartHashmapUniqueCreate(TaskPart):
    snake_name = "hash_map_unique_create"

    def __init__(self, df, expression, dtype, dtype_item, flatten, limit, limit_raise, selection, nthreads, return_inverse):
        expression = str(expression)
        self.nthreads = nthreads
        self.df = df
        # since a df is mutable, store it beforehand
        self.df_fp = self.df.fingerprint()
        self.dtype = dtype
        self.dtype_item = dtype_item
        self.flatten = flatten
        self.expression = expression
        self.limit = limit
        self.limit_raise = limit_raise
        self.selection = selection
        self.return_inverse = return_inverse
        self.chunks = []
        self.values = None

        transient = False
        # TODO: revive non-transient optimization
        # transient = self.df.[str(expression)].transient or self.filtered or self.is_masked(expression)
        # if self.is_string(expression) and not transient:
        #     # string is a special case, only ColumnString are not transient
        #     ar = self.columns[str(expression)]
        #     if not isinstance(ar, ColumnString):
        #         transient = True
        # self.dtype = self.df.data_type(str(expression))
        # self.dtype_item = self.data_type(expression, axis=-1 if flatten else 0)
        # *7 is arbitrary, but we can have more maps than threads to avoid locks
        self.hash_map_unique = vaex.hash.HashMapUnique(self.dtype_item, self.nthreads*7, limit=self.limit)

    def get_bin_count(self):
        return len(self.hash_map_unique)

    @property
    def expressions(self):
        return [self.expression]

    def get_result(self):
        if self.return_inverse:
            return self.hash_map_unique, self.values
        else:
            return self.hash_map_unique

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        ar = blocks[0]
        self._check_row_limit()
        if self.stopped:
            return
        if self.selection:
            selection_mask = selection_masks[0]
            ar = filter(ar, selection_mask)
        if len(ar) == 0:
            return
        if self.dtype.is_list and self.flatten:
            ar = ar.values
        result = self.hash_map_unique.add(ar, return_inverse=self.return_inverse)
        if self.return_inverse:
            values, map_index = result
            self.chunks.append((i1, i2, values, map_index))
        self._check_row_limit()
        if logger.level >= logging.DEBUG:
            logger.debug(f"set uses {sys.getsizeof(self.hash_map_unique):,} bytes (offset {i1:,}, length {i2-i1:,})")
        self._check_row_limit()

    def _check_row_limit(self):
        if self.limit is not None:
            # we only raise when we EXCEED the limit
            if self.limit_raise and len(self.hash_map_unique) > self.limit:
                raise vaex.RowLimitException(f'Resulting hash_map_unique would have >= {self.limit} unique combinations')
            # but we can stop when we are AT the limit
            if not self.limit_raise and len(self.hash_map_unique) >= self.limit:
                self.stopped = True

    def ideal_splits(self, nthreads):
        # TODO, we want to configure this
        return 1 # nthreads

    def reduce(self, others):
        all = [self] + others
        all = [k.hash_map_unique for k in all if k.hash_map_unique is not None]
        hash_map_unique_merged, *others = all
        import time
        t0 = time.time()
        hash_map_unique_merged.merge(others)
        logger.info(f'merge took {time.time()-t0} seconds, size {len(hash_map_unique_merged):,}, byte_size {sys.getsizeof(hash_map_unique_merged):,}')

        if self.return_inverse:
            # sort by row index
            self.chunks.sort(key=lambda x: x[0])
            length = 0
            for i1, i2, values, map_index in self.chunks:
                length += len(values)
            self.values = np.empty(length, vaex.dtype_of(self.chunks[0][2]).numpy)
            # TODO: we could do this parallel, but overhead is small
            for i1, i2, values, map_index in self.chunks:
                hash_map_unique_merged.flatten_values(values, map_index, self.values[i1:i2])

        if self.limit is not None:
            count = len(hash_map_unique_merged)
            if count > self.limit:
                if self.limit_raise:
                    raise vaex.RowLimitException(f'Resulting set has {count:,} unique combinations, which is larger than the allowed value of {self.limit:,}')
                else:
                    hash_map_unique_merged = hash_map_unique_merged.limit(self.limit)
        logger.debug("flatten hashmap...")
        self.hash_map_unique = hash_map_unique_merged.flatten()
        logger.debug("flatten hashmap done")
        self.hash_map_unique._internal.fingerprint = f'hash-map-unique-{self.fingerprint}'

    @classmethod
    def decode(cls, encoding, spec, df, nthreads):
        return cls(df, spec['expression'], encoding.decode('dtype', spec['dtype']), encoding.decode('dtype', spec['dtype_item']),
                   flatten=spec['flatten'], limit=spec['limit'], limit_raise=spec['limit_raise'], selection=spec['selection'], return_inverse=spec['return_inverse'], nthreads=nthreads)

    def memory_usage(self):
        return sys.getsizeof(self.hash_map_unique._internal)


@register
class TaskPartMapReduce(TaskPart):
    snake_name = "map_reduce"

    def __init__(self, df, expressions, map, reduce, converter=lambda x: x, info=False, to_float=False,
                 to_numpy=True, ordered_reduce=False, skip_masked=False, ignore_filter=False, selection=None, pre_filter=False, name="task"):
        TaskPart.__init__(self, df, expressions, name=name, pre_filter=pre_filter)

        self._map = map
        self._reduce = reduce
        self.converter = converter
        self.info = info
        self.ordered_reduce = ordered_reduce
        self.to_float = to_float
        self.to_numpy = to_numpy
        self.skip_masked = skip_masked
        self.ignore_filter = ignore_filter
        if self.pre_filter and self.ignore_filter:
            raise ValueError("Cannot pre filter and also ignore the filter")
        self.selection = selection
        self.values = []

    def get_bin_count(self):
        return 0

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        selection_mask = selection_masks[0]
        if self.to_numpy:
            blocks = [block if isinstance(block, np.ndarray) else block.to_numpy() for block in blocks]
        if self.to_float:
            blocks = [as_flat_float(block) for block in blocks]
        if self.skip_masked:
            masks = [np.ma.getmaskarray(block) for block in blocks if np.ma.isMaskedArray(block)]
            blocks = [block.data if np.ma.isMaskedArray(block) else block for block in blocks]
            mask = None
            if masks:
                # find all 'rows', where all columns are present (not masked)
                mask = masks[0].copy()
                for other in masks[1:]:
                    mask |= other
                blocks = [block[~mask] for block in blocks]

        if not self.ignore_filter:
            selection = self.selection
            if self.pre_filter:
                if selection:
                    blocks = [filter(block, selection_mask) for block in blocks]
            else:
                if selection and self.df.filtered:
                    selection_mask = selection_mask & filter_mask
                    blocks = [filter(block, selection_mask) for block in blocks]
                elif self.df.filtered:
                    blocks = [filter(block, filter_mask) for block in blocks]
                elif selection:
                    blocks = [filter(block, selection_mask) for block in blocks]
        if self.info:
            self.values.append(self._map(thread_index, i1, i2, selection_mask, blocks))
        else:
            self.values.append(self._map(*blocks))

    def reduce(self, others):
        # if self.ordered_reduce:
        #     results.sort(key=lambda x: x[0])
        #     results = [k[1] for k in results]
        # return self.converter(reduce(self._reduce, results))
        values = self.values
        for other in others:
            values.extend(other.values)
        if values:
            self.values = reduce(self._reduce, values)

    def get_result(self):
        return self.values

    @classmethod
    def decode(cls, encoding, spec, df, nthreads):
        spec = spec.copy()
        return cls(df, **spec)


@register
class TaskPartStatistic(TaskPart):
    snake_name = "legacy_statistic"

    def __init__(self, df, shape, expressions, dtype, selections, op, weights, minima, maxima, edges, selection_waslist):
        self.df = df
        self.shape = shape
        self.dtype = dtype
        self.expressions = expressions
        self.op = op
        self.selections = selections
        self.fields = op.fields(weights)
        self.shape_total = (len(self.selections), ) + self.shape + (self.fields,)
        self.grid = np.zeros(self.shape_total, dtype=self.dtype.numpy)
        self.op.init(self.grid)
        self.minima = minima
        self.maxima = maxima
        self.edges = edges
        self.selection_waslist = selection_waslist

    def get_bin_count(self):
        return reduce(operator.mul, self.shape, 1)

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        class Info(object):
            pass
        info = Info()
        info.i1 = i1
        info.i2 = i2
        info.first = i1 == 0
        info.last = i2 == self.df.length_unfiltered()
        info.size = i2 - i1

        masks = [np.ma.getmaskarray(block) for block in blocks if np.ma.isMaskedArray(block)]
        blocks = [block.data if np.ma.isMaskedArray(block) else block for block in blocks]
        blocks = [np.asarray(k) for k in blocks]
        mask = None

        # blocks = [as_flat_float(block) for block in blocks]
        if len(blocks) != 0:
            dtype = np.find_common_type([block.dtype for block in blocks], [])
            if dtype.str in ">f8 <f8 =f8":
                statistic_function = vaex.vaexfast.statisticNd_f8
            elif dtype.str in ">f4 <f4 =f4":
                statistic_function = vaex.vaexfast.statisticNd_f4
            elif dtype.str in ">i8 <i8 =i8":
                dtype = np.dtype(np.float64)
                statistic_function = vaex.vaexfast.statisticNd_f8
            else:
                dtype = np.dtype(np.float32)
                statistic_function = vaex.vaexfast.statisticNd_f4
            # print(dtype, statistic_function, histogram2d)

        if masks:
            mask = masks[0].copy()
            for other in masks[1:]:
                mask |= other
            blocks = [block[~mask] for block in blocks]

        this_thread_grid = self.grid
        for i, selection in enumerate(self.selections):
            if selection:
                selection_mask = selection_masks[i]
                if selection_mask is None:
                    raise ValueError("performing operation on selection while no selection present")
                if mask is not None:
                    selection_mask = selection_mask[~mask]
                selection_blocks = [block[selection_mask] for block in blocks]
            else:
                selection_blocks = [block for block in blocks]
            little_endians = len([k for k in selection_blocks if k.dtype != str and k.dtype.byteorder in ["<", "="]])
            if not ((len(selection_blocks) == little_endians) or little_endians == 0):
                def _to_native(ar):
                    if ar.dtype == str:
                        return ar  # string are always fine
                    if ar.dtype.byteorder not in ["<", "="]:
                        dtype = ar.dtype.newbyteorder()
                        return ar.astype(dtype)
                    else:
                        return ar

                selection_blocks = [_to_native(k) for k in selection_blocks]
            # subblock_weight = None
            subblock_weights = selection_blocks[len(self.expressions):]
            selection_blocks = list(selection_blocks[:len(self.expressions)])
            if len(selection_blocks) == 0 and subblock_weights == []:
                if self.op == vaex.tasks.OP_ADD1:  # special case for counting '*' (i.e. the number of rows)
                    if selection or self.df.filtered:
                        this_thread_grid[i][0] += np.sum(selection_mask)
                    else:
                        this_thread_grid[i][0] += i2 - i1
                else:
                    raise ValueError("Nothing to compute for OP %s" % self.op.code)
            # special case for counting string values etc
            elif len(selection_blocks) == 0 and len(subblock_weights) == 1 and self.op in [vaex.tasks.OP_COUNT]\
                    and (subblock_weights[0].dtype == str or subblock_weights[0].dtype.kind not in 'biuf'):
                weight = subblock_weights[0]
                mask = None
                if weight.dtype != str:
                    if weight.dtype.kind == 'O':
                        mask = vaex.strings.StringArray(weight).mask()
                else:
                    mask = weight.get_mask()
                if selection or self.df.filtered:
                    if mask is not None:
                        this_thread_grid[i][0] += np.sum(~mask)
                    else:
                        this_thread_grid[i][0] += np.sum(selection_mask)
                else:
                    if mask is not None:
                        this_thread_grid[i][0] += len(mask) - mask.sum()
                    else:
                        this_thread_grid[i][0] += len(weight)
            else:
                selection_blocks = [as_flat_array(block, dtype) for block in selection_blocks]
                subblock_weights = [as_flat_array(block, dtype) for block in subblock_weights]
                statistic_function(selection_blocks, subblock_weights, this_thread_grid[i], self.minima, self.maxima, self.op.code, self.edges)
        return i2 - i1
        # return map(self._map, blocks)#[self.map(block) for block in blocks]

    def reduce(self, others):
        grids = [self.grid] + [k.grid for k in others]
        self.grid = self.op.reduce(np.array(grids))
        # If selection was a string, we just return the single selection

    def get_result(self):
        return self.grid if self.selection_waslist else self.grid[0]

    @classmethod
    def decode(cls, encoding, spec, df, nthreads):
        spec = spec.copy()
        spec['op'] = encoding.decode('_op', spec['op'])
        spec['dtype'] = encoding.decode('dtype', spec['dtype'])
        return cls(df, **spec)


@register
class TaskPartAggregation(TaskPart):
    snake_name = "aggregations"

    def __init__(self, df, binners, aggregation_descriptions, dtypes, initial_values=None, nthreads=None):
        self.df = df
        self.has_values = False
        self.dtypes = dtypes
        self.binners = binners
        self.nthreads = nthreads
        # self.expressions_all = expressions
        self.expressions = [binner.expression for binner in binners]
        # TODO: selection and edges in descriptor?
        self.aggregation_descriptions = aggregation_descriptions
        for aggregator_descriptor in self.aggregation_descriptions:
            self.expressions.extend(aggregator_descriptor.expressions)
        # self.expressions = list(set(expressions))
        self.grid = vaex.superagg.Grid([binner.copy() for binner in binners])
        self.nbytes = 0

        def create_aggregator(aggregator_descriptor, selections, initial_values):
            # for each selection, we have a separate aggregator, sharing the grid and binners
            for i, selection in enumerate(selections):
                agg = aggregator_descriptor._create_operation(self.grid, nthreads)
                self.nbytes += sys.getsizeof(agg)
                if initial_values is not None:
                    print(np.asarray(agg))
                    print(initial_values[i])
                    # np.asarray(agg)[:] = initial_values[i]
                    np.copyto(np.asarray(agg), initial_values[i])
                yield agg

        self.aggregations = []
        for i, aggregator_descriptor in enumerate(self.aggregation_descriptions):
            selection = aggregator_descriptor.selection
            selection_waslist = _issequence(selection)
            selections = _ensure_list(selection)
            initial_values_i = initial_values[i] if initial_values else None
            self.aggregations.append((aggregator_descriptor, selections, list(create_aggregator(aggregator_descriptor, selections, initial_values_i)), selection_waslist))

    def get_bin_count(self):
        return reduce(lambda prev, binner: len(binner) * prev, self.binners, 1)

    def memory_usage(self):
        return self.nbytes

    def ideal_splits(self, nthreads):
        return 1

    def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
        # self.check()
        grid = self.grid

        def check_array(x, dtype):
            if vaex.array_types.is_string_type(dtype):
                x = vaex.column._to_string_sequence(x)
            else:
                x = vaex.utils.as_contiguous(x)
                if x.dtype.kind in "mM":
                    # we pass datetime as int
                    x = x.view('uint64')
            return x

        N = i2 - i1
        if filter_mask is not None:
            if blocks:
                N = len(blocks[0])
            else:
                N = filter_mask.sum()
        blocks = [vaex.array_types.to_numpy(block, strict=False) for block in blocks]
        for block in blocks:
            assert len(block) == N, f'Oops, got a block of length {len(block)} while it is expected to be of length {N} (at {i1}-{i2}, filter={filter_mask is not None})'
        block_map = {expr: block for expr, block in zip(self.expressions, blocks)}
        # we need to make sure we keep some objects alive, since the c++ side does not incref
        # on set_data and set_data_mask
        references = []

        def init_binner(binner):
            block = block_map[binner.expression]
            dtype = self.dtypes[binner.expression]
            block = check_array(block, dtype)
            if np.ma.isMaskedArray(block):
                block, mask = block.data, np.ma.getmaskarray(block)
                binner.set_data(thread_index, block)
                binner.set_data_mask(thread_index, mask)
                references.extend([block, mask])
            else:
                binner.set_data(thread_index, block)
                binner.clear_data_mask(thread_index, )
                references.extend([block])

        for binner in grid.binners:
            if hasattr(binner, "binners"):
                for subbinner in binner.binners:
                    init_binner(subbinner)
            else:
                init_binner(binner)
        all_aggregators = []

        selection_index_global = 0
        for agg_desc, selections, aggregation2d, selection_waslist in self.aggregations:
            for selection_index, selection in enumerate(selections):
                agg = aggregation2d[selection_index]
                all_aggregators.append(agg)
                selection_mask = None
                if not (selection is None or selection is False):
                    selection_mask = selection_masks[selection_index_global] # self.df.evaluate_selection_mask(selection, i1=i1, i2=i2, cache=True)  # TODO
                    # TODO: we probably want a way to avoid a to numpy conversion?
                    assert selection_mask is not None
                    selection_mask = np.asarray(selection_mask)
                    references.append(selection_mask)
                    # some aggregators make a distiction between missing value and no value
                    # like nunique, they need to know if they should take the value into account or not
                    if hasattr(agg, 'set_selection_mask'):
                        if selection_mask is not None:
                            agg.set_selection_mask(thread_index, selection_mask)
                        else:
                            agg.clear_selection_mask(thread_index)
                selection_index_global += 1
                if agg_desc.expressions:
                    assert len(agg_desc.expressions) in [1, 2], "only length 1 or 2 supported for now"
                    for i, expression in enumerate(agg_desc.expressions):
                        block = block_map[agg_desc.expressions[i]]
                        dtype = self.dtypes[agg_desc.expressions[i]]
                        # we have data for the aggregator as well
                        if np.ma.isMaskedArray(block):
                            block, mask = block.data, np.ma.getmaskarray(block)
                            block = check_array(block, dtype)
                            agg.set_data(thread_index, block, i)
                            references.extend([block])
                            if selection_mask is None:
                                selection_mask = ~mask
                            else:
                                selection_mask = selection_mask & ~mask
                            references.append(selection_mask)
                        else:
                            block = check_array(block, dtype)
                            agg.set_data(thread_index, block, i)
                            references.extend([block])
                # we only have 1 data mask, since it's locally combined
                if selection_mask is not None:
                    agg.set_data_mask(thread_index, selection_mask)
                    references.extend([selection_mask])
                else:
                    agg.clear_data_mask(thread_index)
        grid.bin(thread_index, all_aggregators, N)
        self.has_values = True

    def reduce(self, others):
        for agg_index, (agg_desc, selections, aggregation, selection_waslist) in enumerate(self.aggregations):
            for selection_index, selection in enumerate(selections):
                agg0 = aggregation[selection_index]
                aggs = [other.aggregations[agg_index][2][selection_index] for other in others]
                agg0.merge(aggs)

    def get_result(self):
        results = []
        for agg_index, (agg_desc, selections, aggregation, selection_waslist) in enumerate(self.aggregations):
            grids = []
            for selection_index, selection in enumerate(selections):
                grid = agg_desc.get_result(aggregation[selection_index])
                grids.append(grid)
            result = np.asarray(grids) if selection_waslist else grids[0]
            if agg_desc.dtype_out.is_primitive or agg_desc.dtype_out.is_temporal:
                dtype_out = agg_desc.dtype_out.to_native()
                result = result.view(dtype_out.numpy)
                result = result.copy()
            results.append(result)
        return results

    def get_values(self):
        values_outer = []
        for agg_index, (agg_desc, selections, aggregation, selection_waslist) in enumerate(self.aggregations):
            values = []
            for selection_index, selection in enumerate(selections):
                agg = aggregation[selection_index]
                values.append(np.asarray(agg))
            values_outer.append(values)
        return values_outer


    @classmethod
    def decode(cls, encoding, spec, df, nthreads):
        # aggs = [vaex.agg._from_spec(agg_spec) for agg_spec in spec['aggregations']]
        aggs = encoding.decode_list('aggregation', spec['aggregations'])
        dtypes = encoding.decode_dict('dtype', spec['dtypes'])
        grid = encoding.decode_list('binner-cpu', spec['binners'], nthreads=nthreads)
        values = encoding.decode_list2('ndarray', spec['values']) if 'values' in spec else None
        # dtypes = {expr: _deserialize_type(type_spec) for expr, type_spec in spec['dtypes'].items()}
        for agg in aggs:
            agg._prepare_types(df)
        return cls(df, grid, aggs, dtypes, initial_values=values, nthreads=nthreads)

    def encode(self, encoding):
        # TODO: get rid of dtypes
        encoded = {
                'grid': encoding.encode('grid', self.grid),
                'aggregations': encoding.encode_list("aggregation", self.aggregation_descriptions),
                'dtypes': encoding.encode_dict("dtype", self.dtypes)
                }
        if self.has_values:
            encoded['values'] = encoding.encode_list2('ndarray', self.get_values())
        return encoded



import os
from pydantic import BaseSettings
import vaex.utils


# in seperate module to avoid circular imports

class ConfigDefault:
    env_file = ".env"
    @classmethod
    def customise_sources(cls, init_settings, env_settings, file_secret_settings):
        return (
            env_settings,
            file_secret_settings,
            init_settings,  # constructor argument last, since they come from global yaml file
        )



"""This module contains struct dtype related expression functionality.

"""

from typing import List

from vaex.utils import _ensure_list
from .expression import Expression
from .dataframe import DataFrame
import functools
import json

import pyarrow as pa
import vaex
from vaex import register_function



class DataFrameAccessorStruct:
    def __init__(self, df):
        self.df : DataFrame = df


    def flatten(self, column=None, recursive=True, join_char='_') -> DataFrame:
        """Returns a DataFrame where each struct column is turned into individual columns.

        Example:
        >>> import vaex
        >>> import pyarrow as pa
        >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"], [3, 4]], names=["col1", "col2", "col3"])
        >>> df = vaex.from_arrays(array=array)
        >>> df
          #  array
          0  {'col1': 1, 'col2': 'a', 'col3': 3}
          1  {'col1': 2, 'col2': 'b', 'col3': 4}
        >>> df.struct.flatten()
          #    array_col1  array_col2      array_col3
          0             1  a                        3
          1             2  b                        4

        :param str or list[str] column: Column or list of columns to use (default is all).
        :param bool recursive: Keep expanding already expanded columns or not.
        """
        df : DataFrame = self.df.copy()
        filter_columns =  set(self.df.get_column_names() if column is None else _ensure_list(column))
        queue : List[str] = self.df.get_column_names()
        column_names = []  # re-order afterwards
        while queue:
            column = queue.pop(0)
            expression : Expression = df[column]
            if column in filter_columns and expression.data_type().is_struct:
                df._hide_column(column)
                # loop in reverse, so the first items ends up at the start of the queue
                for name, projected_expression in reversed(expression.struct.items()):
                    projected_name = f'{column}{join_char}{name}'
                    df[projected_name] = projected_expression
                    queue.insert(0, projected_name)
                    if recursive:
                        filter_columns.add(projected_name)
            else:
                column_names.append(column)
        return df[column_names]

def assert_struct_dtype(struct):
    """Helper function to ensure that struct operations are only applied to struct arrays."""

    try:
        dtype = struct.type  # arrow type
    except AttributeError:
        dtype = struct.dtype  # numpy type

    dtype = vaex.datatype.DataType(dtype)
    if not dtype.is_struct:
        raise TypeError(
            f"Struct functions needs to be applied on struct dtype, got '{dtype}' instead.")


def format_struct_item_vaex_style(struct_item):
    """Provides consistent vaex formatting output for struct items
    handling structs with duplicated field labels for which conversion
    methods (e.g. `as_py()`) are currently (4.0.1) not well supported
    by pyarrow.

    """

    if is_struct_dtype_with_duplicated_field_labels(struct_item.type):
        return format_duplicated_struct_item_vaex_style(struct_item)
    else:
        return struct_item.as_py()


def format_duplicated_struct_item_vaex_style(struct_item):
    """Provides consistent vaex formatting output for struct item with
    duplicated labels.

    """

    mapping = {idx: dtype.name for idx, dtype in enumerate(struct_item.type)}
    values = [f"'{label}': {struct_item[idx].as_py()}"
              for idx, label in mapping.items()]

    return f"{{{', '.join(values)}}}"


def is_struct_dtype_with_duplicated_field_labels(dtype):
    """Check if struct item has duplicated field labels.

    """

    labels = {field.name for field in dtype}
    return len(labels) < dtype.num_fields


def _get_struct_field_label(dtype, identifier):
    """Return the string field label for given field identifier
    which can either be an integer for position based access or
    a string label directly.

    """

    if isinstance(identifier, str):
        return identifier

    return dtype[identifier].name


def assert_struct_dtype_argument(func):
    """Decorator to ensure that struct functions are only applied to expressions containing
    struct dtype. Otherwise, provide helpful error message.

    """

    @functools.wraps(func)
    def wrapper(struct, *args, **kwargs):
        assert_struct_dtype(struct)
        return func(struct, *args, **kwargs)

    return wrapper


def _check_valid_struct_fields(struct, fields):
    """Ensure that fields do exist for given struct and provide helpful error message otherwise.

    """

    # check for existing/valid fields
    valid_field_labels = {x.name for x in struct.type}
    valid_field_indices = set(range(struct.type.num_fields))
    valid_field_lookups = valid_field_labels.union(valid_field_indices)

    non_existant_fields = {field for field in fields
                           if field not in valid_field_lookups}
    if non_existant_fields:
        raise LookupError(
            f"Invalid field lookup provided: {non_existant_fields}. "
            f"Valid field lookups are '{valid_field_lookups}'")

    # check for duplicated field names and provide helpful error message
    labels = [field.name for field in struct.type]
    duplis = {x for x in labels if labels.count(x) > 1}
    duplicated_label_lookup = duplis.intersection(fields)
    if duplicated_label_lookup:
        raise LookupError(f"Invalid field lookup due to duplicated field "
                          f"labels '{duplicated_label_lookup}'. Please use "
                          f"index position based lookup for fields with "
                          f"duplicated labels to uniquely identify relevant "
                          f"field. To get index positions for field labels, "
                          f"please use `{{idx: key for idx, key in enumerate(df.array.struct)}}`.")


@register_function(scope="struct")
@assert_struct_dtype_argument
def struct_get(x, field):
    """Return a single field from a struct array. You may also use the shorthand notation `df.name[:, 'field']`.

    Please note, in case of duplicated field labels, a field can't be uniquely identified. Please
    use index position based access instead. To get corresponding field indices, please use
    `{{idx: key for idx, key in enumerate(df.array.struct)}}`.

    :param {str, int} field: A string (label) or integer (index position) identifying a struct field.
    :returns: an expression containing a struct field.

    Example:

    >>> import vaex
    >>> import pyarrow as pa
    >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"]], names=["col1", "col2"])
    >>> df = vaex.from_arrays(array=array)
    >>> df
      #  array
      0  {'col1': 1, 'col2': 'a'}
      1  {'col1': 2, 'col2': 'b'}

    >>> df.array.struct.get("col1")
    Expression = struct_get(array, 'col1')
    Length: 2 dtype: int64 (expression)
    -----------------------------------
    0  1
    1  2

    >>> df.array.struct.get(0)
    Expression = struct_get(array, 0)
    Length: 2 dtype: int64 (expression)
    -----------------------------------
    0  1
    1  2

    >>> df.array[:, 'col1']
    Expression = struct_get(array, 'col1')
    Length: 2 dtype: int64 (expression)
    -----------------------------------
    0  1
    1  2

    """

    _check_valid_struct_fields(x, [field])
    return x.field(field)


@register_function(scope="struct")
@assert_struct_dtype_argument
def struct_project(x, fields):
    """Project one or more fields of a struct array to a new struct array. You may also use the shorthand notation
    `df.name[:, ['field1', 'field2']]`.

    :param list field: A list of strings (label) or integers (index position) identifying one or more fields.
    :returns: an expression containing a struct array.

    Example:

    >>> import vaex
    >>> import pyarrow as pa
    >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"], [3, 4]], names=["col1", "col2", "col3"])
    >>> df = vaex.from_arrays(array=array)
    >>> df
      #  array
      0  {'col1': 1, 'col2': 'a', 'col3': 3}
      1  {'col1': 2, 'col2': 'b', 'col3': 4}

    >>> df.array.struct.project(["col3", "col1"])
    Expression = struct_project(array, ['col3', 'col1'])
    Length: 2 dtype: struct<col3: int64, col1: int64> (expression)
    --------------------------------------------------------------
    0  {'col3': 3, 'col1': 1}
    1  {'col3': 4, 'col1': 2}

    >>> df.array.struct.project([2, 0])
    Expression = struct_project(array, [2, 0])
    Length: 2 dtype: struct<col3: int64, col1: int64> (expression)
    --------------------------------------------------------------
    0  {'col3': 3, 'col1': 1}
    1  {'col3': 4, 'col1': 2}

    >>> df.array[:, ["col3", "col1"]]
    Expression = struct_project(array, ['col3', 'col1'])
    Length: 2 dtype: struct<col3: int64, col1: int64> (expression)
    --------------------------------------------------------------
    0  {'col3': 3, 'col1': 1}
    1  {'col3': 4, 'col1': 2}

    """

    _check_valid_struct_fields(x, fields)
    arrays = [x.field(field) for field in fields]
    fields = [_get_struct_field_label(x.type, field) for field in fields]

    return pa.StructArray.from_arrays(arrays=arrays, names=fields)



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import asyncio
import threading
import queue
import math
import multiprocessing
import sys
import vaex.utils
import logging
import traceback
import cProfile
import concurrent.futures
import time
import os

from .itertools import buffer

logger = logging.getLogger("vaex.multithreading")

thread_count_default = vaex.utils.get_env_type(int, 'VAEX_NUM_THREADS', multiprocessing.cpu_count())
thread_count_default_io = vaex.utils.get_env_type(int, 'VAEX_NUM_THREADS_IO', thread_count_default + 1)
main_pool = None
main_io_pool = None


thread_pools = {}


def get_thread_pool(nthreads):
    if nthreads not in thread_pools:
        thread_pools[nthreads] = ThreadPoolIndex(nthreads)
    return thread_pools[nthreads]


def get_main_pool():
    global main_pool
    if main_pool is None:
        main_pool = ThreadPoolIndex()
        thread_pools[main_pool.nthreads] = main_pool
    return main_pool


def get_main_io_pool():
    global main_io_pool
    if main_io_pool is None:
        main_io_pool = concurrent.futures.ThreadPoolExecutor(max_workers=thread_count_default_io)
    return main_io_pool





class ThreadPoolIndex(concurrent.futures.ThreadPoolExecutor):
    """Thread pools that adds a thread index as first argument to the callback passed to map

    This is useful if you keep a piece of memory (like ndgrid) per array
    """

    def __init__(self, max_workers=None, *args, **kwargs):
        if max_workers is None:
            max_workers = thread_count_default
        super(ThreadPoolIndex, self).__init__(max_workers, *args, **kwargs)
        self.lock = threading.Lock()
        self.thread_indices = iter(range(1000000))  # enough threads until 2100?
        self.local = threading.local()
        self.nthreads = self._max_workers
        self._debug_sleep = 0

    def map(self, callable, iterator, count, on_error=None, progress=None, cancel=None, unpack=False, use_async=False, **kwargs_extra):
        progress = progress or (lambda x: True)
        cancelled = False

        def wrapped(*args, **kwargs):
            if not cancelled:
                if self.nthreads == 1:
                    self.local.index = 0
                with self.lock:
                    if not hasattr(self.local, 'index'):
                        self.local.index = next(self.thread_indices)
                if unpack:
                    args = args[0]  # it's passed as a tuple.. not sure why
                if self._debug_sleep:
                    # print("SLEEP", self._debug_sleep)
                    time.sleep(self._debug_sleep)
                return callable(self.local.index, *args, **kwargs, **kwargs_extra)
        time_last = time.time() - 100
        min_delta_t = 1. / 10  # max 10 per second
        # we don't want to keep consuming the chunk iterator when we cancel
        chunk_iterator = iterator
        def cancellable_iter():
            for value in chunk_iterator:
                yield value
                if cancelled:
                    break
        if self.nthreads == 1:  # when using 1 thread, it makes debugging easier (better stacktrace)
            if use_async:
                iterator = self._map_async(wrapped, cancellable_iter())
            else:
                iterator = self._map(wrapped, cancellable_iter())
        else:
            if use_async:
                loop = asyncio.get_event_loop()
                iterator = (loop.run_in_executor(self, lambda value=value: wrapped(value)) for value in cancellable_iter())
            else:
                iterator = super(ThreadPoolIndex, self).map(wrapped, cancellable_iter())
        total = 0
        iterator = iter(buffer(iterator, self._max_workers + 3))
        try:
            for value in iterator:
                if use_async:
                    value = yield value
                else:
                    yield value
                if value != None:
                    total += value
                progress_value = (total) / count
                time_now = time.time()
                if progress_value == 1 or (time_now - time_last) > min_delta_t:
                    time_last = time_now
                    if progress(progress_value) is False:
                        cancelled = True
                        cancel()
                        break
        finally:
            if not cancelled:
                cancelled = True
            # consume the rest of the iterators and await them to avoid un-awaited exceptions, which trigger a
            # 'Future exception was never retrieved' printout
            # TODO: since we don't use async any more, I think we can get rid of this
            for value in iterator:
                try:
                    pass
                except:
                    pass

    def _map(self, callable, iterator):
        for i in iterator:
            yield callable(i)

    def _map_async(self, callable, iterator):
        for i in iterator:
            future = asyncio.Future()
            future.set_result(callable(i))
            yield future



"""
Implementation of the dataframe exchange protocol.

Public API
----------

from_dataframe : construct a vaex.dataframe.DataFrame from an input data frame which
                 implements the exchange protocol

For a background and spec, see:
   * https://data-apis.org/blog/dataframe_protocol_rfc/
   * https://data-apis.org/dataframe-protocol/latest/index.html

Notes
-----
- Interpreting a raw pointer (as in ``Buffer.ptr``) is annoying and unsafe to
  do in pure Python. It's more general but definitely less friendly than having
  ``to_arrow`` and ``to_numpy`` methods. So for the buffers which lack
  ``__dlpack__`` (e.g., because the column dtype isn't supported by DLPack),
  this is worth looking at again.
"""
import enum
import collections.abc
import ctypes
from typing import Any, Optional, Tuple, Dict, Iterable, Sequence

import vaex
import pyarrow
import pyarrow as pa
import numpy as np
import pandas as pd


# A typing protocol could be added later to let Mypy validate code using
# `from_dataframe` better.
DataFrameObject = Any
ColumnObject = Any


def from_dataframe_to_vaex(df: DataFrameObject, allow_copy: bool = True) -> vaex.dataframe.DataFrame:
    """
    Construct a vaex DataFrame from ``df`` if it supports ``__dataframe__``
    """
    if isinstance(df, vaex.dataframe.DataFrame):
        return df

    if not hasattr(df, "__dataframe__"):
        raise ValueError("`df` does not support __dataframe__")

    return _from_dataframe_to_vaex(df.__dataframe__(allow_copy=allow_copy))


def _from_dataframe_to_vaex(df: DataFrameObject) -> vaex.dataframe.DataFrame:
    """
    Note: we need to implement/test support for bit/byte masks, chunk handling, etc.
    """
    # Iterate through the chunks
    dataframe = []
    _buffers = []
    for chunk in df.get_chunks():

        # We need a dict of columns here, with each column being an expression.
        columns = dict()
        _k = _DtypeKind
        _buffers_chunks = []  # hold on to buffers, keeps memory alive
        for name in chunk.column_names():
            if not isinstance(name, str):
                raise ValueError(f"Column {name} is not a string")
            if name in columns:
                raise ValueError(f"Column {name} is not unique")

            col = chunk.get_column_by_name(name)
            if col.dtype[0] in (_k.INT, _k.UINT, _k.FLOAT, _k.BOOL):
                # Simple numerical or bool dtype, turn into arrow array
                columns[name], _buf = convert_column_to_ndarray(col)
            elif col.dtype[0] == _k.CATEGORICAL:
                columns[name], _buf = convert_categorical_column(col)
            elif col.dtype[0] == _k.STRING:
                columns[name], _buf = convert_string_column(col)
            else:
                raise NotImplementedError(f"Data type {col.dtype[0]} not handled yet")

            _buffers_chunks.append(_buf)

        dataframe.append(vaex.from_dict(columns))
        # chunk buffers are added to list of all buffers
        _buffers.append(_buffers_chunks)

    if df.num_chunks() == 1:
        _buffers = _buffers[0]

    df_new = vaex.concat(dataframe)
    df_new._buffers = _buffers
    return df_new


class _DtypeKind(enum.IntEnum):
    INT = 0
    UINT = 1
    FLOAT = 2
    BOOL = 20
    STRING = 21  # UTF-8
    DATETIME = 22
    CATEGORICAL = 23


def convert_column_to_ndarray(col: ColumnObject) -> pa.Array:
    """
    Convert an int, uint, float or bool column to an arrow array
    """
    if col.offset != 0:
        raise NotImplementedError("column.offset > 0 not handled yet")

    if col.describe_null[0] not in (0, 1, 3, 4):
        raise NotImplementedError("Null values represented as" "sentinel values not handled yet")

    _buffer, _dtype = col.get_buffers()["data"]
    x = buffer_to_ndarray(_buffer, _dtype)

    # If there are any missing data with mask, apply the mask to the data
    if col.describe_null[0] in (3, 4) and col.null_count > 0:
        mask_buffer, mask_dtype = col._get_validity_buffer()
        mask = buffer_to_ndarray(mask_buffer, mask_dtype)
        x = pa.array(x, mask=mask)
    else:
        x = pa.array(x)

    return x, _buffer


def buffer_to_ndarray(_buffer, _dtype) -> np.ndarray:
    # Handle the dtype
    kind = _dtype[0]
    bitwidth = _dtype[1]
    _k = _DtypeKind
    if _dtype[0] not in (_k.INT, _k.UINT, _k.FLOAT, _k.BOOL):
        raise RuntimeError("Not a boolean, integer or floating-point dtype")

    _ints = {8: np.int8, 16: np.int16, 32: np.int32, 64: np.int64}
    _uints = {8: np.uint8, 16: np.uint16, 32: np.uint32, 64: np.uint64}
    _floats = {32: np.float32, 64: np.float64}
    _np_dtypes = {0: _ints, 1: _uints, 2: _floats, 20: {8: bool}}
    column_dtype = _np_dtypes[kind][bitwidth]

    # No DLPack, so need to construct a new ndarray from the data pointer
    # and size in the buffer plus the dtype on the column
    ctypes_type = np.ctypeslib.as_ctypes_type(column_dtype)
    data_pointer = ctypes.cast(_buffer.ptr, ctypes.POINTER(ctypes_type))

    # NOTE: `x` does not own its memory, so the caller of this function must
    #       either make a copy or hold on to a reference of the column or
    #       buffer! (not done yet, this is pretty awful ...)
    x = np.ctypeslib.as_array(data_pointer, shape=(_buffer.bufsize // (bitwidth // 8),))

    return x


def convert_categorical_column(col: ColumnObject) -> Tuple[pa.DictionaryArray, Any]:
    """
    Convert a categorical column to an arrow dictionary
    """
    catinfo = col.describe_categorical
    if not catinfo["is_dictionary"]:
        raise NotImplementedError("Non-dictionary categoricals not supported yet")
    assert catinfo["categories"] is not None  # sanity check

    if not col.describe_null[0] in (0, 2, 3, 4):
        raise NotImplementedError(
            "Only categorical columns with sentinel "
            "value and masks supported at the moment"
        )

    codes_buffer, codes_dtype = col.get_buffers()["data"]
    codes = buffer_to_ndarray(codes_buffer, codes_dtype)

    if col.describe_null[0] == 2:  # sentinel value
        sentinel = [col.describe_null[1]] * col.size
        mask = codes == sentinel
        indices = pa.array(codes, mask=mask)
    elif col.describe_null[0] in (3, 4) and col.null_count > 0:  # masked missing values
        mask_buffer, mask_dtype = col._get_validity_buffer()
        mask = buffer_to_ndarray(mask_buffer, mask_dtype)
        indices = pa.array(codes, mask=mask)
    else:
        indices = pa.array(codes)

    labels_buffer, labels_dtype = catinfo["categories"].get_buffers()["data"]
    if labels_dtype[0]  == _DtypeKind.STRING:
        labels, _ = convert_string_column(catinfo["categories"])
    else:
        labels = buffer_to_ndarray(labels_buffer, labels_dtype)

    values = pa.DictionaryArray.from_arrays(indices, labels)

    return values, codes_buffer


def convert_string_column(col: ColumnObject) -> Tuple[pa.Array, list]:
    """
    Convert a string column to a Arrow array.
    """
    # Missing
    if col.null_count > 0:
        if col.describe_null != (3, 0):
            raise TypeError("Only support arrow style mask data")

    # Retrieve the data buffers
    buffers = col.get_buffers()

    dbuffer, bdtype = buffers["data"]  # buffer containing the UTF-8 code units
    obuffer, odtype = buffers["offsets"]  #  buffer containing the index offsets demarcating the beginning and end of each string
    mbuffer, mdtype = buffers["validity"]  # buffer indicating the presence of missing values

    # Convert the buffers to NumPy arrays
    dt = (_DtypeKind.UINT, 8, None, None)  # note: in order to go from STRING to an equivalent ndarray, we claim that the buffer is uint8 (i.e., a byte array)
    dbuf = buffer_to_ndarray(dbuffer, dt)

    obuf = buffer_to_ndarray(obuffer, odtype)
    mbuf = buffer_to_ndarray(mbuffer, mdtype)

    # not sure what the best way to communicate the two types of strings is
    if obuffer._x.dtype == "int64":
        arrow_type = pa.large_utf8()
    elif obuffer._x.dtype == "int32":
        arrow_type = pa.utf8()
    else:
        raise TypeError(f"oops")
    length = obuf.size - 1

    buffers = [None, pa.py_buffer(obuf), pa.py_buffer(dbuf)]
    arrow_array = pa.Array.from_buffers(arrow_type, length, buffers)

    # Apply the mask
    if col.null_count > 0:
        arrow_array = pa.array(arrow_array.tolist(), mask=mbuf)

    return arrow_array, buffers


# Implementation of interchange protocol
# --------------------------------------


class _VaexBuffer:
    """
    Data in the buffer is guaranteed to be contiguous in memory.
    """

    def __init__(self, x: np.ndarray, allow_copy: bool = True) -> None:
        """
        Handle only regular columns (= numpy arrays) for now.
        """
        if not x.strides == (x.dtype.itemsize,):
            # The protocol does not support strided buffers, so a copy is
            # necessary. If that's not allowed, we need to raise an exception.
            if allow_copy:
                x = x.copy()
            else:
                raise RuntimeError("Exports cannot be zero-copy in the case " "of a non-contiguous buffer")

        # Store the numpy array in which the data resides as a private
        # attribute, so we can use it to retrieve the public attributes
        self._x = x

    @property
    def bufsize(self) -> int:
        """
        Buffer size in bytes
        """
        return self._x.size * self._x.dtype.itemsize

    @property
    def ptr(self) -> int:
        """
        Pointer to start of the buffer as an integer
        """
        return self._x.__array_interface__["data"][0]

    def __dlpack__(self):
        """
        DLPack not implemented in Vaex, so leave it out here
        """
        raise NotImplementedError("__dlpack__")

    def __dlpack_device__(self) -> Tuple[enum.IntEnum, int]:
        """
        Device type and device ID for where the data in the buffer resides.
        """

        class Device(enum.IntEnum):
            CPU = 1

        return (Device.CPU, None)

    def __repr__(self) -> str:
        return "VaexBuffer(" + str({"bufsize": self.bufsize, "ptr": self.ptr, "device": self.__dlpack_device__()[0].name}) + ")"


class _VaexColumn:
    """
    A column object, with only the methods and properties required by the
    interchange protocol defined.

    A column can contain one or more chunks. Each chunk can contain up to three
    buffers - a data buffer, a mask buffer (depending on null representation),
    and an offsets buffer (if variable-size binary; e.g., variable-length
    strings).

    Note: this Column object can only be produced by ``__dataframe__``, so
          doesn't need its own version or ``__column__`` protocol.

    """

    def __init__(self, column: vaex.expression.Expression, allow_copy: bool = True) -> None:
        """
        Note: assuming column is an expression.
        The values of an expression can be NumPy or Arrow.
        """
        if not isinstance(column, vaex.expression.Expression):
            raise NotImplementedError("Columns of type {} not handled " "yet".format(type(column)))

        # Store the column as a private attribute
        self._col = column
        self._allow_copy = allow_copy

    def size(self) -> int:
        """
        Size of the column, in elements.

        Corresponds to DataFrame.num_rows() if column is a single chunk;
        equal to size of this current chunk otherwise.

        Is a method rather than a property because it may cause a (potentially
        expensive) computation for some dataframe implementations.
        """
        return int(len(self._col.df))

    @property
    def offset(self) -> int:
        """
        Offset of first element. Always zero.
        """
        return 0

    @property
    def dtype(self) -> Tuple[enum.IntEnum, int, str, str]:
        """
        Dtype description as a tuple ``(kind, bit-width, format string, endianness)``

        Kind :

            - INT = 0
            - UINT = 1
            - FLOAT = 2
            - BOOL = 20
            - STRING = 21   # UTF-8
            - DATETIME = 22
            - CATEGORICAL = 23

        Bit-width : the number of bits as an integer
        Format string : data type description format string in Apache Arrow C
                        Data Interface format.
        Endianness : current only native endianness (``=``) is supported

        Notes:

            - Kind specifiers are aligned with DLPack where possible (hence the
              jump to 20, leave enough room for future extension)
            - Masks must be specified as boolean with either bit width 1 (for bit
              masks) or 8 (for byte masks).
            - Dtype width in bits was preferred over bytes
            - Endianness isn't too useful, but included now in case in the future
              we need to support non-native endianness
            - Went with Apache Arrow format strings over NumPy format strings
              because they're more complete from a dataframe perspective
            - Format strings are mostly useful for datetime specification, and
              for categoricals.
            - For categoricals, the format string describes the type of the
              categorical in the data buffer. In case of a separate encoding of
              the categorical (e.g. an integer to string mapping), this can
              be derived from ``self.describe_categorical``.
            - Data types not included: complex, Arrow-style null, binary, decimal,
              and nested (list, struct, map, union) dtypes.
        """
        dtype = self._col.dtype

        # Categorical
        # If it is internal, kind must be categorical (23)
        # If it is external (call from_dataframe), dtype must give type of the data
        if self._col.df.is_category(self._col):
            return (_DtypeKind.CATEGORICAL, 64, "u", "=")  # what should be the default??

        return self._dtype_from_vaexdtype(dtype)

    def _dtype_from_vaexdtype(self, dtype) -> Tuple[enum.IntEnum, int, str, str]:
        """
        See `self.dtype` for details
        """
        # Note: 'c' (complex) not handled yet (not in array spec v1).
        #       'b', 'B' (bytes), 'S', 'a', (old-style string) 'V' (void) not handled
        #       datetime, timedelta not implemented yet
        _k = _DtypeKind
        _np_kinds = {"i": _k.INT, "u": _k.UINT, "f": _k.FLOAT, "b": _k.BOOL, "U": _k.STRING, "M": _k.DATETIME, "m": _k.DATETIME}
        kind = _np_kinds.get(dtype.kind, None)

        if kind is None:
            # Check if it's a an Arrow dictonary
            if not isinstance(self._col.values, np.ndarray) and isinstance(self._col.values.type, pa.DictionaryType):
                kind = 23
            # Check if it's a string dtype (dtype.kind == 'O')
            elif dtype == "string":
                kind = 21
            else:
                raise ValueError(f"Data type {dtype} not supported by exchange protocol")

        if kind not in (_k.INT, _k.UINT, _k.FLOAT, _k.BOOL, _k.CATEGORICAL, _k.STRING):
            raise NotImplementedError(f"Data type {dtype} not handled yet")

        bitwidth = dtype.numpy.itemsize * 8
        if not isinstance(self._col.values, np.ndarray) and isinstance(self._col.values.type, pa.DictionaryType):
            format_str = self._col.index_values().dtype.numpy.str
        else:
            format_str = dtype.numpy.str
        endianness = dtype.byteorder if not kind == _k.CATEGORICAL else "="
        return (kind, bitwidth, format_str, endianness)

    @property
    def describe_categorical(self) -> Dict[str, Any]:
        """
        If the dtype is categorical, there are two options:
        - There are only values in the data buffer.
        - There is a separate non-categorical Column encoding categorical values.

        Raises TypeError if the dtype is not categorical

        Returns the dictionary with description on how to interpret the data buffer:
            - "is_ordered" : bool, whether the ordering of dictionary indices is
                             semantically meaningful.
            - "is_dictionary" : bool, whether a mapping of
                                categorical values to other objects exists
            - "categories" : Column representing the (implicit) mapping of indices to
                             category values (e.g. an array of cat1, cat2, ...).
                             None if not a dictionary-style categorical.

        TBD: are there any other in-memory representations that are needed?
        """
        if not self.dtype[0] == _DtypeKind.CATEGORICAL:
            raise TypeError(
                "describe_categorical only works on a column with "
                "categorical dtype!"
            )
        df = vaex.from_dict({"labels": self._col.df.category_labels(self._col)})
        labels = df["labels"]
        categories = _VaexColumn(labels)
        return {
            "is_ordered": False,
            "is_dictionary": True,
            "categories": categories,
        }

    @property
    def describe_null(self) -> Tuple[int, Any]:
        """
        Return the missing value (or "null") representation the column dtype
        uses, as a tuple ``(kind, value)``.

        Kind:

            - 0 : non-nullable
            - 1 : NaN/NaT
            - 2 : sentinel value
            - 3 : bit mask
            - 4 : byte mask

        Value : if kind is "sentinel value", the actual value.  If kind is a bit
        mask or a byte mask, the value (0 or 1) indicating a missing value. None
        otherwise.
        """
        _k = _DtypeKind
        kind = self.dtype[0]
        value = None
        if kind in (_k.INT, _k.UINT, _k.FLOAT, _k.BOOL, _k.STRING):
            if self._col.dtype.is_arrow:
                # arrow arrays always allow for null values
                # where 0 encodes a null/missing value
                null = 3
                value = 0
            elif self._col.is_masked:
                # masked arrays are always numpy.ma arrays
                null = 4
                value = 1
            else:
                # otherwise we have a normal numpy array
                null = 0
                value = None
        elif kind == _k.CATEGORICAL:
            if self._col.dtype.is_arrow:
                # arrow dictionary allows for null values
                # where 0 encodes a null/missing value
                null = 3
                value = 0
            else:
                # otherwise we have categorize() with a normal numpy array
                null = 0
                value = None
        else:
            raise NotImplementedError(f"Data type {self.dtype} not yet supported")

        return null, value

    @property
    def null_count(self) -> int:
        """
        Number of null elements. Should always be known.
        """
        # Whilst vaex-proper does not treat all NA values as null (i.e. NaNs),
        # the interchange protocol does, meaning we have to break away from vaex
        # semantics here and return a count of all NA values.
        # See https://github.com/vaexio/vaex/issues/2120
        return self._col.countna()

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Store specific metadata of the column.
        """
        return {}

    def num_chunks(self) -> int:
        """
        Return the number of chunks the column consists of.
        """
        if isinstance(self._col.values, pa.ChunkedArray):
            return self._col.values.num_chunks
        else:
            return 1

    def get_chunks(self, n_chunks: Optional[int] = None) -> Iterable["_VaexColumn"]:
        """
        Return an iterator yielding the chunks.

        See `DataFrame.get_chunks` for details on ``n_chunks``.
        """
        if n_chunks == None:
            size = self.size()
            n_chunks = self.num_chunks()
            i = self._col.df.evaluate_iterator(self._col, chunk_size=size // n_chunks)
            iterator = []
            for i1, i2, chunk in i:
                iterator.append(_VaexColumn(self._col[i1:i2]))
            return iterator
        elif self.num_chunks == 1:
            size = self.size()
            i = self._col.df.evaluate_iterator(self._col, chunk_size=size // n_chunks)
            iterator = []
            for i1, i2, chunk in i:
                iterator.append(_VaexColumn(self._col[i1:i2]))
            return iterator

        else:
            raise ValueError(f"Column {self._col.expression} is already chunked.")

    def get_buffers(self) -> Dict[str, Any]:
        """
        Return a dictionary containing the underlying buffers.

        The returned dictionary has the following contents:

            - "data": a two-element tuple whose first element is a buffer
                      containing the data and whose second element is the data
                      buffer's associated dtype.
            - "validity": a two-element tuple whose first element is a buffer
                          containing mask values indicating missing data and
                          whose second element is the mask value buffer's
                          associated dtype. None if the null representation is
                          not a bit or byte mask.
            - "offsets": a two-element tuple whose first element is a buffer
                         containing the offset values for variable-size binary
                         data (e.g., variable-length strings) and whose second
                         element is the offsets buffer's associated dtype. None
                         if the data buffer does not have an associated offsets
                         buffer.
        """
        buffers = {}
        buffers["data"] = self._get_data_buffer()
        try:
            buffers["validity"] = self._get_validity_buffer()
        except:
            buffers["validity"] = None

        try:
            buffers["offsets"] = self._get_offsets_buffer()
        except:
            buffers["offsets"] = None

        return buffers

    def _get_data_buffer(self) -> Tuple[_VaexBuffer, Any]:  # Any is for self.dtype tuple
        """
        Return the buffer containing the data and the buffer's associated dtype.
        """
        _k = _DtypeKind
        if self.dtype[0] in (_k.INT, _k.UINT, _k.FLOAT, _k.BOOL):
            # If arrow array is boolean .to_numpy changes values for some reason
            # For that reason data is transferred to numpy through .tolist
            if self.dtype[0] == _k.BOOL and isinstance(self._col.values, (pa.Array, pa.ChunkedArray)):
                buffer = _VaexBuffer(np.array(self._col.tolist(), dtype=bool))
            else:
                buffer = _VaexBuffer(self._col.to_numpy())
            dtype = self.dtype
        elif self.dtype[0] == _k.CATEGORICAL:
            # TODO: Use expression.codes (https://github.com/vaexio/vaex/pull/1503), when merged
            if isinstance(self._col.values, (pa.DictionaryArray)):
                buffer = _VaexBuffer(self._col.index_values().to_numpy())
                dtype = self._dtype_from_vaexdtype(self._col.index_values().dtype)
            else:
                codes = self._col.values
                indices = self._col
                offset = self._col.df.category_offset(self._col)
                if offset:
                    indices -= offset
                buffer = _VaexBuffer(indices.to_numpy())
                dtype = self._dtype_from_vaexdtype(self._col.dtype)
        elif self.dtype[0] == _k.STRING:
            bitmap_buffer, offsets, string_bytes = self._col.evaluate().buffers()

            if string_bytes is None:
                string_bytes = np.array([], dtype="uint8")
            else:
                string_bytes = np.frombuffer(string_bytes, "uint8", len(string_bytes))
            buffer = _VaexBuffer(string_bytes)

            # Define the dtype for the returned buffer
            dtype = (_k.STRING, 8, "u", "=")  # note: currently only support native endianness
        else:
            raise NotImplementedError(f"Data type {self._col.dtype} not handled yet")

        return buffer, dtype

    def _get_validity_buffer(self) -> Tuple[_VaexBuffer, Any]:
        """
        Return the buffer containing the mask values indicating missing data and
        the buffer's associated dtype.

        Raises RuntimeError if null representation is not a bit or byte mask.
        """
        null, invalid = self.describe_null

        _k = _DtypeKind
        if null == 3 or null == 4:  # arrow
            mask = self._col.ismissing()

            # if arrow use .tolist and then turn it into np.array
            # if numpy masked turn the mask into numpy
            data = np.array(mask.tolist()) if null == 3 else mask.to_numpy()

            buffer = _VaexBuffer(data)
            dtype = self._dtype_from_vaexdtype(mask.dtype)

            return buffer, dtype

        elif null == 0:
            msg = "This column is non-nullable so does not have a mask"
        elif null == 1:
            msg = "This column uses NaN as null so does not have a separate mask"
        else:
            raise NotImplementedError("See self.describe_null")

        raise RuntimeError(msg)

    def _get_offsets_buffer(self) -> Tuple[_VaexBuffer, Any]:
        """
        Return the buffer containing the offset values for variable-size binary
        data (e.g., variable-length strings) and the buffer's associated dtype.

        Raises RuntimeError if the data buffer does not have an associated
        offsets buffer.
        """
        _k = _DtypeKind
        if self.dtype[0] == _k.STRING:
            bitmap_buffer, offsets, string_bytes = self._col.evaluate().buffers()

            if self._col.evaluate().type == pyarrow.string():
                offsets = np.frombuffer(offsets, np.int32, len(offsets) // 4)
                dtype = (_k.INT, 32, "i", "=")
            else:
                offsets = np.frombuffer(offsets, np.int64, len(offsets) // 8)
                dtype = (_k.INT, 64, "l", "=")
            buffer = _VaexBuffer(offsets)

        else:
            raise RuntimeError("This column has a fixed-length dtype so does not have an offsets buffer")

        return buffer, dtype


class _VaexDataFrame:
    """
    A data frame class, with only the methods required by the interchange
    protocol defined.

    Instances of this (private) class are returned from
    ``vaex.dataframe.DataFrame.__dataframe__`` as objects with the methods and
    attributes defined on this class.
    """

    def __init__(self, df: vaex.dataframe.DataFrame, nan_as_null: bool = False, allow_copy: bool = True) -> None:
        """
        Constructor - an instance of this (private) class is returned from
        `vaex.dataframe.DataFrame.__dataframe__`.
        """
        self._df = df
        # ``nan_as_null`` is a keyword intended for the consumer to tell the
        # producer to overwrite null values in the data with ``NaN`` (or ``NaT``).
        # This currently has no effect; once support for nullable extension
        # dtypes is added, this value should be propagated to columns.
        self._nan_as_null = nan_as_null
        self._allow_copy = allow_copy

    def __dataframe__(self, nan_as_null : bool = False, allow_copy : bool = True) -> "_VaexDataFrame":
        return _VaexDataFrame(self._df, nan_as_null=nan_as_null, allow_copy=allow_copy)

    @property
    def metadata(self) -> Dict[str, Any]:
        return {}

    def num_columns(self) -> int:
        return len(self._df.columns)

    def num_rows(self) -> int:
        return len(self._df)

    def num_chunks(self) -> int:
        if isinstance(self.get_column(0)._col.values, pa.ChunkedArray):
            return self.get_column(0)._col.values.num_chunks
        else:
            return 1

    def column_names(self) -> Iterable[str]:
        return self._df.get_column_names()

    def get_column(self, i: int) -> _VaexColumn:
        return _VaexColumn(self._df[:, i], allow_copy=self._allow_copy)

    def get_column_by_name(self, name: str) -> _VaexColumn:
        return _VaexColumn(self._df[name], allow_copy=self._allow_copy)

    def get_columns(self) -> Iterable[_VaexColumn]:
        return [_VaexColumn(self._df[name], allow_copy=self._allow_copy) for name in self._df.columns]

    def select_columns(self, indices: Sequence[int]) -> "_VaexDataFrame":
        if not isinstance(indices, collections.Sequence):
            raise ValueError("`indices` is not a sequence")

        names = []
        for i in indices:
            names.append(self._df[:, i].expression)

        return self.select_columns_by_name(names)

    def select_columns_by_name(self, names: Sequence[str]) -> "_VaexDataFrame":
        if not isinstance(names, collections.Sequence):
            raise ValueError("`names` is not a sequence")

        return _VaexDataFrame(self._df[names], allow_copy=self._allow_copy)

    def get_chunks(self, n_chunks: Optional[int] = None) -> Iterable["_VaexDataFrame"]:
        """
        Return an iterator yielding the chunks.
        TODO: details on ``n_chunks``
        """
        n_chunks = n_chunks if n_chunks is not None else self.num_chunks()
        size = self.num_rows()
        chunk_size = (size + n_chunks - 1) // n_chunks
        column_names = self.column_names()
        i = self._df._future().evaluate_iterator(column_names, chunk_size=chunk_size)
        for i1, i2, chunk in i:
            yield _VaexDataFrame(vaex.from_items(*zip(column_names, chunk)))

